from ..type import SimpleType
from asn1PERser.classes.templates.creator import template_filler
from asn1PERser.classes.module import already_filled_template


class EnumeratedType(SimpleType):
    def __init__(self):
        super(EnumeratedType, self).__init__()
        self._EnumerationItems = []
        self.typereference = self.__class__.__name__

    def fill_template(self, has_parent=False):
        root_named_values = []
        addition_named_values = []
        extension_marker_present = False
        enumeration_index = 0
        for EnumerationItem in self.EnumerationItems:
            if EnumerationItem == '...':
                extension_marker_present = True
                continue
            if not extension_marker_present:
                root_named_values.append({'field_name': str(EnumerationItem).strip(),
                                          'field_index': enumeration_index})
            else:
                addition_named_values.append({'field_name': str(EnumerationItem).strip(),
                                              'field_index': enumeration_index})
            enumeration_index += 1
        filled_template = template_filler.fill(asn_type=self.__class__.__name__,
                                               class_name=self.template_class_name,
                                               class_type=self.__class__.__name__,
                                               extension_marker=extension_marker_present,
                                               root_named_values=root_named_values,
                                               addition_named_values=addition_named_values)
        if has_parent:
            return (None, filled_template)
        if filled_template in already_filled_template:
            return ''
        already_filled_template.add(filled_template)
        return filled_template

    @property
    def EnumerationItems(self):
        return self._EnumerationItems

    @EnumerationItems.setter
    def EnumerationItems(self, EnumerationItem):
        if isinstance(EnumerationItem, str):
            self._EnumerationItems.append(EnumerationItem)
        else:
            self._EnumerationItems.append({EnumerationItem[1]: EnumerationItem[0]})

    def parse_value(self, value):
        return "'" + value + "'"

    def __repr__(self):
        return '\n\t'.join([super(EnumeratedType, self).__repr__()] + \
                           ['{' + str(item).strip() + '}' for item in self.EnumerationItems if item])



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


maxNumber = IntegerType(255)


class MyInt1(IntegerType):
    subtypeSpec = ValueRange(-5, -1)


class MyInt2(IntegerType):
    subtypeSpec = ValueRange(-25, 20000000, extensionMarker=True)


class MyInt3(IntegerType):
    subtypeSpec = ValueRange(-25, MAX)


class MyInt4(IntegerType):
    subtypeSpec = ValueRange(MIN, 0)


class MyInt5(IntegerType):
    subtypeSpec = ValueRange(MIN, maxNumber)


class MyOctetString1(OctetStringType):
    subtypeSpec = ValueSize(1, 4)


class MyOctetString2(OctetStringType):
    subtypeSpec = ValueSize(0, 30, extensionMarker=True)


class MyOctetString3(OctetStringType):
    subtypeSpec = ValueSize(0, maxNumber, extensionMarker=True)


class MyBitString1(BitStringType):
    subtypeSpec = ValueSize(10, 100)


class MyBitString2(BitStringType):
    subtypeSpec = ValueSize(100, 10000, extensionMarker=True)


class MyBitString3(BitStringType):
    subtypeSpec = ValueSize(0, maxNumber, extensionMarker=True)


class MySeqOf1(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, maxNumber)
    componentType = OctetStringType()


class MySeqOf2(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, maxNumber, extensionMarker=True)
    componentType = MyBitString1()





from pyasn1.type.univ import BitString
from asn1PERser.codec.per.encoder import encode_bitstring
from asn1PERser.codec.per.decoder import decode_bitstring
from asn1PERser.classes.types.constraint import NoConstraint


class BitStringType(BitString):
    subtypeSpec = NoConstraint()

    def fill_field_list(self, field_list):
        bitstring_field_list = encode_bitstring(self)
        if bitstring_field_list:
            field_list.extend(bitstring_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_bitstring(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        return int(self.asBinary(), 2)



from pyparsing import Or, And, Forward, Literal, ZeroOrMore, Group
from asn1PERser.classes.parse_actions import parse_ModuleDefinition
from .assignment_def import TypeAssignment, ValueAssignment
from .common_def import DefinedValue
from .lexical_items import COMMA, LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, \
    assignment, typereference, valuereference, empty, modulereference, \
    number, identifier


FROM = Literal("FROM")
EXPLICIT_TAGS = Literal("EXPLICIT TAGS")
IMPLICIT_TAGS = Literal("IMPLICIT TAGS")
AUTOMATIC_TAGS = Literal("AUTOMATIC TAGS")
DEFINITIONS = Literal("DEFINITIONS").setResultsName("DEFINITIONS")
BEGIN = Literal("BEGIN").setResultsName("BEGIN")
END = Literal("END").setResultsName("END")


# 13.1
Assignment = Or([
    TypeAssignment,
    ValueAssignment,
    # XMLValueAssignment,
    # ValueSetTypeAssignment,
    # ObjectClassAssignment,    # Rec. ITU-T X.681 | ISO/IEC 8824-2, 9.1
    # ObjectAssignment,         # Rec. ITU-T X.681 | ISO/IEC 8824-2, 11.1
    # ObjectSetAssignment,      # Rec. ITU-T X.681 | ISO/IEC 8824-2, 12.1
    # ParameterizedAssignment   # Rec. ITU-T X.683 | ISO/IEC 8824-4, 8.1
])


# 13.1
# AssignmentList = Forward()
# AssignmentList << Or([
#                       Assignment,
#                       AssignmentList + Assignment
# ])
AssignmentList = ZeroOrMore(Assignment)


# 13.1
Reference = Or([
    typereference,
    valuereference,
    # objectclassreference,
    # objectreference,
    # objectsetreference
])


# 13.1
Symbol = Or([
    Reference,
    # ParameterizedReference
])


# 13.1
SymbolList = Forward()
SymbolList << Or([
    Symbol,
    SymbolList + COMMA + Symbol
])


# 13.1
AssignedIdentifier = Or([
    # ObjectIdentifierValue,
    DefinedValue,
    empty
])


# 13.1
GlobalModuleReference = modulereference + AssignedIdentifier


# 13.1
SymbolsFromModule = SymbolList + FROM + GlobalModuleReference


# 13.1
SymbolsFromModuleList = Or([
    SymbolsFromModule,
    # SymbolsFromModuleList + SymbolsFromModule
])


# 13.1
SymbolsImported = Or([
    SymbolsFromModuleList,
    empty
])


# 13.1
Imports = Or([
    # "IMPORTS" + SymbolsImported + ";",
    empty
])


# 13.1
SymbolsExported = Or([
    SymbolList,
    empty
])


# 13.1
Exports = Or([
    # "EXPORTS" + SymbolsExported + ";",
    # "EXPORTS ALL" + ";",
    empty
])


# 13.1
ModuleBody = Or([
    Exports + Imports + AssignmentList,
    empty
])

# 13.1
ExtensionDefault = Or([
    # "EXTENSIBILITY IMPLIED",
    empty
])


# 13.1
TagDefault = Or([
    # EXPLICIT_TAGS,
    # IMPLICIT_TAGS,
    AUTOMATIC_TAGS,
    empty
])


# 13.1
EncodingReferenceDefault = Or([
    # encodingreference + "INSTRUCTIONS",
    empty
])


# 13.1
DefinitiveNumberForm = number


# 13.1
DefinitiveNameAndNumberForm = identifier + LEFT_CURLY_BRACKET + DefinitiveNumberForm + RIGHT_CURLY_BRACKET


# 13.1
DefinitiveObjIdComponent = Or([
    # NameForm,
    DefinitiveNumberForm,
    DefinitiveNameAndNumberForm
])


# 13.1
# DefinitiveObjIdComponentList = Or([
#                                    DefinitiveObjIdComponent,
#                                    DefinitiveObjIdComponent + DefinitiveObjIdComponentList
#                                  ])


# 13.1
# DefinitiveOIDandIRI = And([
#                            DefinitiveOID,
#                            IRIValue
#                          ])


# 13.1
# DefinitiveOID = "{" + DefinitiveObjIdComponentList + "}"


# 13.1
DefinitiveIdentification = Or([
    # DefinitiveOID,
    # DefinitiveOIDandIRI,
    empty
])


# 13.1
ModuleIdentifier = And([
    modulereference,
    # DefinitiveIdentification
])


# 13.1
ModuleDefinition = And([
    ModuleIdentifier.setResultsName("ModuleIdentifier"),
    DEFINITIONS,
    EncodingReferenceDefault.setResultsName("EncodingReferenceDefault"),
    TagDefault.setResultsName("TagDefault"),
    ExtensionDefault.setResultsName("ExtensionDefault"),
    assignment,
    BEGIN,
    Group(ModuleBody).setResultsName("ModuleBody"),
    # EncodingControlSection + \
    END
]).setParseAction(parse_ModuleDefinition)



from pyasn1.type.univ import Integer
from pyasn1.type.univ import noValue
from asn1PERser.codec.per.encoder import encode_integer
from asn1PERser.codec.per.decoder import decode_integer
from asn1PERser.classes.types.constraint import NoConstraint


class IntegerType(Integer):
    subtypeSpec = NoConstraint()

    def __init__(self, value=noValue, **kwargs):
        super(IntegerType, self).__init__(value, **kwargs)

    def fill_field_list(self, field_list):
        integer_field_list = encode_integer(self)
        if integer_field_list:
            field_list.extend(integer_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_integer(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        return int(self)



from copy import deepcopy
from collections import namedtuple
from asn1PERser.classes.types.builtin import SequenceType, DefinedType, \
    VisibleString, NumericString, IntegerType, DateType, RealType, BooleanType, \
    ChoiceType, EnumeratedType, IA5String, UTF8String, OctetStringType, \
    BitStringType, SequenceOfType, builtin_types
from asn1PERser.classes.module import ModuleDefinition, typereference_to_type
from asn1PERser.asn_definitions.constraint_def import Constraint


constraint_tuple = namedtuple('constraint', 'lowerEndpoint upperEndpoint extensionMarker')


def parse_ModuleDefinition(toks):
    module = ModuleDefinition()
    module.ModuleIdentifier = toks['ModuleIdentifier'][0]
    module.EncodingReferenceDefault = toks['EncodingReferenceDefault'] if 'EncodingReferenceDefault' in toks.keys() else ''
    module.TagDefault = toks['TagDefault'] if 'TagDefault' in toks.keys() else ''
    module.ExtensionDefault = toks['ExtensionDefault'] if 'ExtensionDefault' in toks.keys() else ''
    module.ModuleBody = toks['ModuleBody'] if 'ModuleBody' in toks.keys() else ''
    return module


def parse_ValueAssignment(toks):
    item_valuereference = toks[0]
    item_type = toks[1][0]
    item_value = toks[-1]
    item_type.valuereference = item_valuereference
    item_type.Value = item_value
    return item_type


def parse_TypeAssignment(toks):
    item_typereference = toks[0]
    item_type = toks[2][0]

    if item_type.typereference in builtin_types:
        item_type.typereference = item_typereference
    else:
        item_type.identifier = item_typereference
    typereference_to_type[item_typereference] = deepcopy(item_type)
    return item_type


def parse_BitStringType(toks):
    return BitStringType()


def parse_OctetStringType(toks):
    return OctetStringType()


def parse_SequenceType(toks):
    sequence_type = SequenceType()
    for Type in toks[1:]:
        sequence_type.ComponentTypeList = Type
    return sequence_type


def parse_SequenceOfType(toks):
    sequence_of_type = SequenceOfType()
    component_type = toks[1][0]
    sequence_of_type.ComponentType = component_type
    return sequence_of_type


def parse_ChoiceType(toks):
    choice_type = ChoiceType()
    for Type in toks[1:]:
        choice_type.AlternativeTypeList = Type
    return choice_type


def parse_DefinedType(toks):
    return DefinedType(toks[0])


def parse_IA5String(toks):
    return IA5String()


def parse_UTF8String(toks):
    return UTF8String()


def parse_VisibleString(toks):
    return VisibleString()


def parse_NumericString(toks):
    return NumericString()


def parse_IntegerType(toks):
    return IntegerType()


def parse_RealType(toks):
    return RealType()


def parse_BooleanType(toks):
    return BooleanType()


def parse_ComponentType(toks):
    identifier = toks[0]
    Type = toks[1][0]
    Type.identifier = identifier
    try:
        if toks[2] == 'OPTIONAL':
            Type.optional = True
        if toks[2] == 'DEFAULT':
            Type.default = toks[3]
    except IndexError:
        pass
    return Type


def parse_NamedType(toks):
    pass


def parse_ConstrainedType(toks):
    Type = toks[0]
    constraint = toks[1]
    Type.constraint = constraint
    return Type


def parse_TypeWithConstraint(toks):
    constrain = toks[1]
    if toks[0] == "SEQUENCE":
        componentType = toks[3]
        Type = parse_SequenceOfType(["SEQUENCE OF", componentType])
    if toks[0] == "SET":
        raise NotImplemented
    return [Type, constrain]


def parse_DateType(toks):
    return DateType()


def parse_AlternativeTypeList(toks):
    alternate_type_list = []
    for identifier, Type in zip(toks[0::2], toks[1::2]):
        Type = Type[0]
        Type.identifier = identifier
        alternate_type_list.append(Type)
    return alternate_type_list


def parse_ExtensionAdditionAlternativesList(toks):
    toks = list(toks)
    alternate_type_list = []
    is_extension_addition_group = False
    if '[[' in toks:
        is_extension_addition_group = True
        toks.remove('[[')
        toks.remove(']]')
    for identifier, Type in zip(toks[0::2], toks[1::2]):
        Type = Type[0]
        Type.identifier = identifier
        alternate_type_list.append(Type)
    if is_extension_addition_group:
        alternate_type_list.insert(0, '[[')
        alternate_type_list.append(']]')
    return alternate_type_list


def parse_EnumeratedType(toks):
    enumerated_type = EnumeratedType()
    for EnumerationItem in toks[1:]:
        enumerated_type.EnumerationItems = EnumerationItem
    return enumerated_type


def parse_Constraint(toks):
    constraint = toks[0]
    if constraint[0] == 'SIZE':
        return constraint[1]
    constraint = '(' + ' '.join(constraint) + ')'
    parsed = list(Constraint.parseString(constraint))
    ext_mark = False
    if '...' in parsed:
        parsed.remove('...')
        ext_mark = True
    return constraint_tuple(lowerEndpoint=parsed[0], upperEndpoint=parsed[-1], extensionMarker=ext_mark)



from pyasn1.type.univ import Choice, noValue
from asn1PERser.codec.per.encoder import encode_choice
from asn1PERser.codec.per.decoder import decode_choice
from asn1PERser.classes.types.constraint import NoConstraint


class ChoiceType(Choice):
    subtypeSpec = NoConstraint()
    rootComponent = None
    extensionAddition = None
    extensionAdditionGroups = []

    def fill_field_list(self, field_list):
        choice_field_list = encode_choice(self)
        if choice_field_list:
            field_list.extend(choice_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_choice(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        for componentType, componentValue in self.items():
            if componentValue is not noValue and componentValue.isValue:
                value_dict = componentValue.to_dict(is_root=False)
                if value_dict is not noValue:
                    return {componentType: value_dict}
                return {}



import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_sequence_of import SCHEMA_constrained_seq_of_no_extension, \
    SCHEMA_constrained_seq_of_extension_present, SCHEMA_no_constrains_sequence_of, DATA_seq_of


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_no_constrains_sequence_of(), '00',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=False, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_no_constrains_sequence_of(), '0103008707',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_no_constrains_sequence_of(), '020300870702FCB3',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_no_constrains_sequence_of(), '030300870702FCB30117',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False)),
    (SCHEMA_no_constrains_sequence_of(), '040300870702FCB3011703030D40',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False)),
    (SCHEMA_no_constrains_sequence_of(), '050300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
])
def test_no_constrains_sequence_of_integer_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), '4003008707',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), '800300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), '000300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), '400300870702FCB30117',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), '800300870702FCB3011703030D40',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), 'C00300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
])
def test_constrained_sequence_of_no_extension_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_seq_of_no_extension(lb=1, ub=1), '03008707',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=1), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=2, ub=2), '0300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=3, ub=3), '0300870702FCB30117',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=3, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=4, ub=4), '0300870702FCB3011703030D40',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=4), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False)),
    (SCHEMA_constrained_seq_of_no_extension(lb=5, ub=5), '0300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=5, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
])
def test_constrained_sequence_of_of_fixed_length_no_extension_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), '2003008707',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), '400300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), '000300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), '200300870702FCB30117',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), '400300870702FCB3011703030D40',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), '600300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
])
def test_constrained_sequence_of_with_extension_and_num_of_elems_is_within_extension_root_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), '800103008707',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), '80020300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), '80030300870702FCB30117',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), '80020300870702FCB3',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), '80050300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), '80040300870702FCB3011703030D40',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False)),
    (SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), '80050300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), '80050300870702FCB3011703030D400105',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), '800103008707',
        DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
])
def test_constrained_sequence_of_with_extension_and_num_of_elems_is_not_within_extension_root_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


minNumber = IntegerType(-10)


maxNumber = IntegerType(255)


class MyBitString(BitStringType):
    subtypeSpec = ValueSize(1, maxNumber, extensionMarker=True)


class TcpData(IntegerType):
    pass


class SrcUdpTput(IntegerType):
    pass


class TgtUdpTput(IntegerType):
    pass


class InnerSequence(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('i0', IntegerType()),
        NamedType('i1', IntegerType()),
    )
    componentType = rootComponent


class MySeq1(SequenceType):
    class data(ChoiceType):
        class my_int_2(IntegerType):
            subtypeSpec = ValueRange(1, 2)

        rootComponent = AdditiveNamedTypes(
            NamedType('inner-seq', InnerSequence()),
            NamedType('my-int-1', IntegerType()),
            NamedType('my-int-2', my_int_2()),
            NamedType('my-bit', MyBitString()),
        )
        componentType = rootComponent

    class my_int_3(IntegerType):
        subtypeSpec = ValueRange(10, 20)

    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
        NamedType('my-int-3', my_int_3()),
        NamedType('my-octet', OctetStringType()),
    )
    componentType = rootComponent


class MySeq2(SequenceType):
    class enum_one(EnumeratedType):
        subtypeSpec = ExtensionMarker(True)
        enumerationRoot = NamedValues(
            ('one', 0),
            ('two', 1),
        )
        namedValues = enumerationRoot

    class o1(OctetStringType):
        subtypeSpec = ValueSize(1, 10)

    rootComponent = AdditiveNamedTypes(
        NamedType('i0', IntegerType()),
        NamedType('enum-one', enum_one()),
        NamedType('o1', o1()),
    )
    componentType = rootComponent


class SrcTcpData(SequenceType):
    class data(SequenceOfType):
        subtypeSpec = SequenceOfValueSize(1, maxNumber)
        componentType = TcpData()

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
    )
    componentType = rootComponent


class TgtTcpData(SequenceType):
    class data(SequenceOfType):
        subtypeSpec = SequenceOfValueSize(1, maxNumber)
        componentType = TcpData()

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
    )
    componentType = rootComponent


class SrcUdpData(SequenceType):
    class data(SequenceOfType):
        subtypeSpec = SequenceOfValueSize(1, maxNumber)
        componentType = SrcUdpTput()

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
    )
    componentType = rootComponent


class TgtUdpData(SequenceType):
    class data(SequenceOfType):
        subtypeSpec = SequenceOfValueSize(1, maxNumber)
        componentType = TgtUdpTput()

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
    )
    componentType = rootComponent


class BigIpData(SequenceType):
    class d0(IntegerType):
        subtypeSpec = ValueRange(-1, 1)

    class d1(OctetStringType):
        subtypeSpec = ValueSize(1, 20)

    rootComponent = AdditiveNamedTypes(
        NamedType('d0', d0()),
        NamedType('d1', d1()),
    )
    componentType = rootComponent


class IpData(SequenceType):
    class data(SequenceOfType):
        subtypeSpec = SequenceOfValueSize(1, maxNumber)
        componentType = BigIpData()

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('data', data()),
    )
    componentType = rootComponent


class MyChoice(ChoiceType):
    class enum_two(EnumeratedType):
        enumerationRoot = NamedValues(
            ('three', 0),
            ('four', 1),
        )
        namedValues = enumerationRoot

    class inner_seq_2(SequenceType):
        class i1(BitStringType):
            subtypeSpec = ValueSize(1, 2)

        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', i1()),
        )
        componentType = rootComponent

    class o1(OctetStringType):
        subtypeSpec = ValueSize(1, 10)

    rootComponent = AdditiveNamedTypes(
        NamedType('enum-two', enum_two()),
        NamedType('inner-seq-2', inner_seq_2()),
        NamedType('o1', o1()),
        NamedType('my-int-4', IntegerType()),
    )
    componentType = rootComponent





import os
import logging
import logging.config


def logging_setup():
    logging.config.fileConfig(os.path.join(os.path.dirname(__file__), 'logging.cfg'))
    logging.getLogger('asn1perser').addHandler(logging.NullHandler())



from asn1PERser.codec.per.decoder import decode
from asn1PERser.test.parsing.asn1_python_code.SimpleProtocol import SimpleMessage
from asn1PERser.test.per.encoder.test_per_encode_SimpleProtocol import DATA_alive_message, DATA_start_message, \
    DATA_data_message_1, DATA_data_message_2


def test_alive_message_can_be_decoded():
    assert decode(per_stream=bytearray.fromhex('4C5CD5FE55403039'), asn1Spec=SimpleMessage()) == DATA_alive_message()


def test_start_message_can_be_decoded():
    assert decode(per_stream=bytearray.fromhex('00000AC05CD5FE55403039D8ED0000'), asn1Spec=SimpleMessage()) == DATA_start_message()


def test_data_message_1_can_be_decoded():
    assert decode(per_stream=bytearray.fromhex('70D90320F0F0F880DEAD40BEEF40FEED00AAC0BBBBBBBB'), asn1Spec=SimpleMessage()) == DATA_data_message_1()


def test_data_message_2_can_be_decoded():
    assert decode(per_stream=bytearray.fromhex('60010043DEADBEEF'), asn1Spec=SimpleMessage()) == DATA_data_message_2()



import pytest
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.BooleanType import BooleanType


@pytest.mark.parametrize("boolean, encoded", [
    (BooleanType(value=True), '80'),
    (BooleanType(value=False), '00')
])
def test_boolean_type_can_be_encoded(boolean, encoded):
    assert per_encoder(boolean) == bytearray.fromhex(encoded)



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class DataInt(IntegerType):
    pass


class OctetName(OctetStringType):
    pass


class MySeqOf(SequenceOfType):
    componentType = IntegerType()


class MyInnerSeq(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('one', IntegerType()),
        NamedType('two', MySeqOf()),
    )
    componentType = rootComponent


class DataSeq(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('myBool0', BooleanType()),
        NamedType('myBitString', BitStringType()),
        NamedType('myInnerSeq', MyInnerSeq()),
    )
    componentType = rootComponent


class MySeq(SequenceType):
    class some_enum(EnumeratedType):
        enumerationRoot = NamedValues(
            ('one', 0),
            ('two', 1),
            ('three', 2),
        )
        namedValues = enumerationRoot

    class some_seqOf(SequenceOfType):
        componentType = OctetName()

    rootComponent = AdditiveNamedTypes(
        NamedType('some-seq', DataSeq()),
        NamedType('some-enum', some_enum()),
        NamedType('some-integer', DataInt()),
        NamedType('some-seqOf', some_seqOf()),
    )
    componentType = rootComponent


class Record_One(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('value', IntegerType()),
    )
    componentType = rootComponent


class Record_List(SequenceOfType):
    componentType = Record_One()


class MySeq_Two(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('records', Record_List()),
    )
    componentType = rootComponent





import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.classes.data.builtin.BooleanType import BooleanType


@pytest.mark.parametrize("schema, encoded, value", [
    (BooleanType, '80', BooleanType(value=True)),
    (BooleanType, '00', BooleanType(value=False)),
])
def test_boolean_type_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



from pyparsing import Or, Group, ZeroOrMore, Literal
from .lexical_items import number, HYPHEN_MINUS, identifier, LEFT_PARENTHESIS, RIGHT_PARENTHESIS, \
    COMMA, modulereference, FULL_STOP, valuereference, RIGHT_CURLY_BRACKET, LEFT_CURLY_BRACKET, \
    cstring


NULL = Literal("NULL")


# 19.1
SignedNumber = Or([
    number,
    HYPHEN_MINUS + number
]).setParseAction(lambda toks: ''.join(toks))


# 14.6
ExternalValueReference = modulereference + FULL_STOP + valuereference


# 14.1
DefinedValue = Or([
    ExternalValueReference,
    valuereference,
    # ParameterizedValue
])


# 19.1
NamedNumber = Or([
    Group(identifier + LEFT_PARENTHESIS + SignedNumber + RIGHT_PARENTHESIS),
    Group(identifier + LEFT_PARENTHESIS + DefinedValue + RIGHT_PARENTHESIS)
])


# 19.1
NamedNumberList = Or([
    NamedNumber + ZeroOrMore(COMMA + NamedNumber),
    # NamedNumber,
    # NamedNumberList + COMMA + NamedNumber    # ORIGINAL
])


# 41.8
TableRow = number


# 41.8
TableColumn = number


# 41.8
Tuple = LEFT_CURLY_BRACKET + TableColumn + COMMA + TableRow + RIGHT_CURLY_BRACKET


# 41.8
Cell = number


# 41.8
Row = number


# 41.8
Plane = number


# 41.8
Group = number


# 41.8
Quadruple = LEFT_CURLY_BRACKET + Group + COMMA + Plane + COMMA + Row + COMMA + Cell + RIGHT_CURLY_BRACKET


# 41.8
CharsDefn = Or([
    cstring,
    Quadruple,
    Tuple,
    DefinedValue
])


# 41.8
CharSyms = Or([
    CharsDefn + ZeroOrMore(COMMA + CharsDefn),
    # CharsDefn,
    # CharSyms + COMMA + CharsDefn    # ORIGINAL
])


# 41.8
CharacterStringList = LEFT_CURLY_BRACKET + CharSyms + RIGHT_CURLY_BRACKET




from collections import OrderedDict
from pyasn1.type.univ import Sequence, noValue
from asn1PERser.codec.per.encoder import encode_sequence
from asn1PERser.codec.per.decoder import decode_sequence
from asn1PERser.classes.types.constraint import NoConstraint


class SequenceType(Sequence):
    subtypeSpec = NoConstraint()
    rootComponent = None
    extensionAddition = None
    extensionAdditionGroups = []

    @property
    def isValue(self):
        base_isValue = super(SequenceType, self).isValue
        if base_isValue:
            return base_isValue
        if self.components is noValue:
            return False
        for component in self.components:
            if component is not noValue and component.isValue:
                return True
        return False

    def fill_field_list(self, field_list):
        sequence_field_list = encode_sequence(self)
        if sequence_field_list:
            field_list.extend(sequence_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_sequence(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        component_dict = OrderedDict()
        for componentType, componentValue in self.items():
            if (componentValue is not noValue and componentValue.isValue) or hasattr(componentValue, 'componentType'):
                value_dict = componentValue.to_dict(is_root=False)
                if value_dict is not noValue:
                    component_dict[componentType] = value_dict

        if is_root:
            return component_dict
        if not component_dict:
            return noValue
        return component_dict



import pytest
from pyasn1.type.namedtype import NamedType, NamedTypes, DefaultedNamedType, OptionalNamedType
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.OctetStringType import OctetStringType
from asn1PERser.classes.types.constraint import ValueSize


def SCHEMA_no_constrain_octetstring():
    class MyOctetString(OctetStringType):
        '''
        MyBitString ::= OCTET STRING
        '''
        pass
    return MyOctetString


def SCHEMA_constrained_octetstring(lb, ub):
    class MyOctetString(OctetStringType):
        '''
        MyBitString ::= OCTET STRING (SIZE(lowerEndpoint..upperEndpoint))
        '''
        subtypeSpec = ValueSize(lb, ub)
    return MyOctetString


def SCHEMA_constrained_ext_octetstring(lb, ub):
    class MyOctetString(OctetStringType):
        '''
        MyBitString ::= OCTET STRING (SIZE(lowerEndpoint..upperEndpoint, ...))
        '''
        subtypeSpec = ValueSize(lb, ub, extensionMarker=True)
    return MyOctetString


def DATA_octetstring(schema, bin_val=None, hex_val=None):
    if bin_val is not None:
        octetstring = schema(binValue=bin_val)
    elif hex_val is not None:
        octetstring = schema(hexValue=hex_val)
    return octetstring


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1'), '0180'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='0'), '0100'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='01'), '0140'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='10'), '0180'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1111'), '01F0'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='111100'), '01F0'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='1111000011110'), '02F0F0'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='0000110000110011'), '020C33'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), bin_val='01011100101100110101101101101110001110010101110010101'),
     '075CB35B6E395CA8'),
])
def test_no_constrain_octetstring_as_binary_value_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='0'), '0100'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='1'), '0110'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='01'), '0101'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='10'), '0110'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='00A'), '0200A0'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='AA0'), '02AA00'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='ABCDEFA'), '04ABCDEFA0'),
    (DATA_octetstring(SCHEMA_no_constrain_octetstring(), hex_val='0123ABCDEFA0123'), '080123ABCDEFA01230'),
])
def test_no_constrain_octetstring_as_hex_value_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    # (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=0), bin_val=''), '00'),    # does not work
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=0), hex_val=''), '00'),
])
def test_constrained_octetstring_of_zero_length_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='1'), '80'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='01110'), '70'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), bin_val='00001111'), '0F'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), hex_val='C'), 'C0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=1, ub=1), hex_val='CF'), 'CF'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='111111110'), 'FF00'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='0000111100001'), '0F08'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), bin_val='1100110011001100'), 'CCCC'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='ABC'), 'ABC0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='000'), '0000'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='0000'), '0000'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=2, ub=2), hex_val='FFFF'), 'FFFF'),
])
def test_constrained_same_len_octetstring_with_len_le_2_octetes_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=3, ub=3), bin_val='000000010000000100000001'), '010101'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=3, ub=3), hex_val='ABCDEF'), 'ABCDEF'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=4, ub=4), bin_val='01111111111111111111111100000000'), '7FFFFF00'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=4, ub=4), hex_val='ABCDEF01'), 'ABCDEF01'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=17, ub=17), hex_val='11223344556677889900AABBCCDDEEFF11'),
     '11223344556677889900AABBCCDDEEFF11'),
])
def test_constrained_same_len_octetstring_with_len_gt_2_and_lt_64K_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val=''), '00'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='A'), '10A0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='A0'), '10A0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=15), hex_val='112233445566778899001122334455'),
     'F0112233445566778899001122334455'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='A'), '08A0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='A0'), '08A0'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='112233445566778899001122334455'),
     '78112233445566778899001122334455'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=0, ub=22), hex_val='11223344556677889900112233445566778899001122'),
     'B011223344556677889900112233445566778899001122'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=15), hex_val='112233445566778899AA'),
     '00112233445566778899AA'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=15), hex_val='112233445566778899AABBCCDDEEFF'),
    'A0112233445566778899AABBCCDDEEFF'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=22), hex_val='112233445566778899AA'),
     '00112233445566778899AA'),
    (DATA_octetstring(SCHEMA_constrained_octetstring(lb=10, ub=22), hex_val='112233445566778899AABBCCDDEEFF001122334455'),
    'B0112233445566778899AABBCCDDEEFF001122334455'),
])
def test_constrained_different_len_octet_string_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val=''), '00'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=1, ub=1), hex_val='A'), '5000'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=1, ub=1), hex_val='AB'), '5580'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=5, ub=5), hex_val='112233445'), '001122334450'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=5, ub=5), hex_val='1122334455'), '001122334455'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=15, ub=15), hex_val='112233445566778899001122334455'),
     '00112233445566778899001122334455'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=22, ub=22), hex_val='11223344556677889900112233445566778899001122'),
     '0011223344556677889900112233445566778899001122'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val=''), '00'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='A'), '20A0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='AB'), '20AB'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='ABC'), '40ABC0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=2), hex_val='ABCD'), '40ABCD'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='A'), '08A0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='AB'), '08AB'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='112233445566778899'), '48112233445566778899'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=15), hex_val='112233445566778899AABBCCDDEEFF'),
     '78112233445566778899AABBCCDDEEFF'),
])
def test_constrained_with_extension_when_len_is_within_extension_root_octetstring_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("octetstring, encoded", [
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='A'), '8001A0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='AB'), '8001AB'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=0, ub=0), hex_val='1122334455'), '80051122334455'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val=''), '8000'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='A'), '8001A0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='AB'), '8001AB'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='112233'), '8003112233'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=2, ub=2), hex_val='1122334455'), '80051122334455'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val=''), '8000'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='A'), '8001A0'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='AB'), '8001AB'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='1122'), '80021122'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='1122334455'), '80051122334455'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=3, ub=4), hex_val='11223344556677'), '800711223344556677'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=10, ub=22), hex_val='112233445566778899'), '8009112233445566778899'),
    (DATA_octetstring(SCHEMA_constrained_ext_octetstring(lb=10, ub=22), hex_val='1122334455667788990011223344556677889900112233'),
     '80171122334455667788990011223344556677889900112233'),
])
def test_constrained_with_extension_when_len_is_NOT_within_extension_root_octetstring_can_be_encoded(octetstring, encoded):
    assert per_encoder(octetstring) == bytearray.fromhex(encoded)



import os
from jinja2 import Environment, FileSystemLoader


class TemplateFiller(object):
    def __init__(self):
        file_loader = FileSystemLoader(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'files/'))
        self.env = Environment(loader=file_loader)

    def fill(self, asn_type, type_or_value='type', **kwargs):
        if asn_type in ('BooleanType', 'IntegerType', 'OctetStringType', 'BitStringType') and type_or_value == 'value':
            template = self.env.get_template('simple_value.txt')
        else:
            template = self.env.get_template('{}.txt'.format(asn_type))
        return template.render(**kwargs)


template_filler = TemplateFiller()



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


constant_Int = IntegerType(10)


constant_Bool = BooleanType(True)


constant_Octet_String_one = OctetStringType(hexValue='DEAD')


constant_Octet_String_two = OctetStringType(binValue='11110000')


constant_Bit_String_one = BitStringType(hexValue='BEEF')


constant_Bit_String_two = BitStringType(binValue='01010101')


class MyInt(IntegerType):
    pass


class My_Int(IntegerType):
    pass


class MyBool(BooleanType):
    pass


class My_Bool(BooleanType):
    pass


class MyOctetString(OctetStringType):
    pass


class My_Octet_String(OctetStringType):
    pass


class MyBitString(BitStringType):
    pass


class My_Bit_String(BitStringType):
    pass


class MyEnumerated(EnumeratedType):
    enumerationRoot = NamedValues(
        ('one', 0),
        ('two', 1),
        ('three', 2),
    )
    namedValues = enumerationRoot


class My_Enumerated(EnumeratedType):
    enumerationRoot = NamedValues(
        ('one', 0),
        ('two', 1),
        ('three', 2),
    )
    namedValues = enumerationRoot


my_constant_Int = MyInt(200)


my_constant_Int_Two = My_Int(12345)


my_constant_Bool = MyBool(False)


my_constant_Bool_Two = My_Bool(True)


my_constant_Octet_String_one = MyOctetString(hexValue='FEED')


my_constant_Octet_String_one_Two = My_Octet_String(hexValue='DEEF')


my_constant_Octet_String_two = MyOctetString(binValue='00001111')


my_constant_Octet_String_two_Two = My_Octet_String(binValue='11110000')


my_constant_Bit_String_one = MyBitString(hexValue='ABCD')


my_constant_Bit_String_one_Two = My_Bit_String(hexValue='DCBA')


my_constant_Bit_String_two = MyBitString(binValue='10101010')


my_constant_Bit_String_two_Two = My_Bit_String(binValue='01010101')





import binascii
import logging
import codecs
from collections import deque
from .encoder import bit_field_size, get_bit_field_size, integer_range, get_range_octet_len, \
    is_constrained_whole_number, is_semi_constrained_whole_number, is_constant


__all__ = ['decode']

logger = logging.getLogger("asn1perser.decoder")


class Offset(object):
    def __init__(self, octets=0, bits=0):
        self.prev_octets = octets
        self.prev_bits = bits
        self.curr_octets = octets
        self.curr_bits = bits
        logger.debug("Bit pointer initialized. Current position: (%s octets, %s bits)", self.curr_octets, self.curr_bits)

    def __add__(self, other):
        self.prev_octets = self.curr_octets
        self.prev_bits = self.curr_bits
        self.curr_octets += other.octets
        self.curr_bits += other.bits
        if self.curr_bits >= 8:
            self.curr_octets += self.curr_bits // 8
            self.curr_bits = self.curr_bits - 8 * (self.curr_bits // 8)
        logger.debug("Bit pointer position: (%s octets, %s bits)", self.curr_octets, self.curr_bits)
        return self

    def align(self):
        bits_to_align = ((1 + self.curr_bits // 8) * 8 - self.curr_bits) if self.curr_bits > 0 else 0
        if bits_to_align:
            self.prev_octets = self.curr_octets
            self.prev_bits = 0
            self.curr_octets += (self.curr_bits // 8 if self.curr_bits % 8 == 0 else (1 + self.curr_bits // 8))
            self.curr_bits = 0
        logger.debug("Aligned %s bits. Current bit pointer position: (%s octets, %s bits)",
                      bits_to_align, self.curr_octets, self.curr_bits)


class PerBytes(object):
    def __init__(self, byte_string):
        self.offset = Offset()
        self.binary_string = bin(int(byte_string, 16))[2:].zfill(len(byte_string) * 4)

    def next(self, size=bit_field_size(octets=1, bits=0), octet_align=False):
        if octet_align:
            self.offset.align()
        bits = self[self.offset:self.offset + size]
        logger.debug("Bits taken: '%s'", 'none' if (size.octets == 0 and size.bits == 0) else "0b" + bits)
        val = int(bits, 2)
        return val

    def __getitem__(self, item):
        bits_start = (item.start.prev_octets * 8) + item.start.prev_bits
        bits_stop = (item.stop.curr_octets * 8) + item.stop.curr_bits
        if bits_start == bits_stop:
            return '0'
        return self.binary_string[bits_start:bits_stop]

    def __len__(self):
        return int(len(self.binary_string) / 8)


def decode(per_stream, asn1Spec, log_level=logging.WARNING):
    logger.setLevel(log_level)
    per_bytes = PerBytes(binascii.hexlify(per_stream))
    logger.debug("Decoding of bytes: '%s' started. Binary representation: '%s'", codecs.encode(per_stream, "hex_codec"), per_bytes.binary_string)
    decoded = asn1Spec.create_field_list(per_bytes)
    if decoded == '':
        return asn1Spec.__class__('')
    return decoded


def decode_boolean(boolean, per_bytes):    # 12
    logger.info("Decoding boolean: %s", boolean.__class__.__name__)
    boolean_val = per_bytes.next(size=bit_field_size(octets=0, bits=1))
    logger.info("%s: value: %s", boolean.__class__.__name__, boolean_val)
    logger.info("%s: decoding finished.", boolean.__class__.__name__)
    return boolean.__class__(bool(boolean_val))


def decode_integer(integer, per_bytes):
    logger.info("Decoding integer: %s", integer.__class__.__name__)
    if integer.subtypeSpec.extensionMarker:    # 13.1
        logger.info("%s: extension marker should be present", integer.__class__.__name__)
        decoded_value_not_within_extension_root_bit = per_bytes.next(size=bit_field_size(octets=0, bits=1))
        logger.info("%s: is value within extension range: %s", integer.__class__.__name__, decoded_value_not_within_extension_root_bit)
        if decoded_value_not_within_extension_root_bit:
            unconstrained_whole_number = decode_unconstrained_whole_number(per_bytes)
            logger.info("%s: value: %s", integer.__class__.__name__, unconstrained_whole_number)
            logger.info("%s: decoding finished.", integer.__class__.__name__)
            return integer.__class__(unconstrained_whole_number)
        else:
            constrained_whole_number = decode_constrained_whole_number_from_field_list(per_bytes=per_bytes,
                                                                                       lowerEndpoint=integer.subtypeSpec.lowerEndpoint,
                                                                                       upperEndpoint=integer.subtypeSpec.upperEndpoint)
            logger.info("%s: value: %s", integer.__class__.__name__, constrained_whole_number)
            logger.info("%s: decoding finished.", integer.__class__.__name__)
            return integer.__class__(constrained_whole_number)
    else:
        if is_constant(integer):
            logger.error("%s: constant integer is not supported")
            raise NotImplemented
        elif is_constrained_whole_number(integer):
            if integer.subtypeSpec.lowerEndpoint == integer.subtypeSpec.upperEndpoint:    # 13.2.1
                logger.info("%s: is contrained to single value: %s", integer.__class__.__name__, integer.subtypeSpec.lowerEndpoint)
                logger.info("%s: decoding finished.", integer.__class__.__name__)
                return (integer.__class__(integer.subtypeSpec.lowerEndpoint))
            else:    # 13.2.2
                logger.info("%s: is contrained between %s and %s. Decoding value...", integer.__class__.__name__, integer.subtypeSpec.lowerEndpoint, integer.subtypeSpec.upperEndpoint)
                constrained_whole_number = decode_constrained_whole_number_from_field_list(per_bytes=per_bytes,
                                                                                           lowerEndpoint=integer.subtypeSpec.lowerEndpoint,
                                                                                           upperEndpoint=integer.subtypeSpec.upperEndpoint)
                logger.info("%s: value: %s", integer.__class__.__name__, constrained_whole_number)
                logger.info("%s: decoding finished.", integer.__class__.__name__)
                return integer.__class__(constrained_whole_number)
        elif is_semi_constrained_whole_number(integer):
            logger.info("%s: is semi contrained - minimum value: %s", integer.__class__.__name__, integer.subtypeSpec.lowerEndpoint)
            semi_constrained_whole_number = decode_semi_constrained_whole_number_from_field_list(per_bytes=per_bytes,
                                                                                                 lowerBound=integer.subtypeSpec.lowerEndpoint)
            logger.info("%s: value: %s", integer.__class__.__name__, semi_constrained_whole_number)
            logger.info("%s: decoding finished.", integer.__class__.__name__)
            return integer.__class__(semi_constrained_whole_number)
        else:    # 13.2.4
            logger.info("%s: is unconstrained whole number. Decoding value...", integer.__class__.__name__)
            unconstrained_whole_number = decode_unconstrained_whole_number(per_bytes)
            logger.info("%s: value: %s", integer.__class__.__name__, unconstrained_whole_number)
            logger.info("%s: decoding finished.", integer.__class__.__name__)
            return integer.__class__(unconstrained_whole_number)


def decode_enumerated(enumerated, per_bytes):
    logger.info("Decoding enumerated: %s", enumerated.__class__.__name__)
    if not enumerated.subtypeSpec.extensionMarker:    # 14.2
        enumeration_index = decode_constrained_whole_number(per_bytes, lowerBound=0, upperBound=len(list(enumerated.namedValues)) - 1)
        logger.info("%s: index: %s", enumerated.__class__.__name__, enumeration_index)
        logger.info("%s: decoding finished.", enumerated.__class__.__name__)
        return enumerated.__class__(enumerated.namedValues.getName(enumeration_index))
    else:    # 14.3
        logger.info("%s: extension marker should be present", enumerated.__class__.__name__)
        value_is_extension_addition = per_bytes.next(size=bit_field_size(octets=0, bits=1))
        logger.info("%s: is index within extension range: %s", enumerated.__class__.__name__, value_is_extension_addition)
        if not value_is_extension_addition:
            enumeration_index = decode_constrained_whole_number(per_bytes, lowerBound=0,
                                                                upperBound=len(list(enumerated.enumerationRoot)) - 1)
            logger.info("%s: index: %s", enumerated.__class__.__name__, enumeration_index)
            logger.info("%s: decoding finished.", enumerated.__class__.__name__)
            return enumerated.__class__(enumerated.namedValues.getName(enumeration_index))
        else:
            enumeration_index = decode_normally_small_non_negative_whole_number(per_bytes, lower_bound=0)
            enumeration_index += len(enumerated.enumerationRoot)
            logger.info("%s: index: %s", enumerated.__class__.__name__, enumeration_index)
            logger.info("%s: decoding finished.", enumerated.__class__.__name__)
            return enumerated.__class__(enumerated.namedValues.getName(enumeration_index))


def decode_bitstring(bitstring, per_bytes):
    logger.info("Decoding bitstring: %s", bitstring.__class__.__name__)
    if bitstring.subtypeSpec.extensionMarker:    # 16.6
        logger.info("%s: extension marker should be present", bitstring.__class__.__name__)
        extension_bit_field = per_bytes.next(size=bit_field_size(octets=0, bits=1))
        logger.info("%s: is bitstring value extened: %s", bitstring.__class__.__name__, extension_bit_field)
        if not extension_bit_field:
            pass
        else:
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            bitstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                     constrained=False,
                                                                     n=per_bytes,
                                                                     lowerBound=0,
                                                                     upperBound=None)
            bitstring_val = per_bytes.next(size=bit_field_size(octets=0, bits=bitstring_length_determinant))
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            bitstring_val = bin(bitstring_val)[2:].zfill(bitstring_length_determinant)
            logger.info("%s: value: %s", bitstring.__class__.__name__, bitstring_val)
            logger.info("%s: decoding finished.", bitstring.__class__.__name__)
            return bitstring.__class__(bitstring_val)
    if bitstring.subtypeSpec.upperEndpoint == 0:    # 16.8
        logger.info("%s: upperEndpoint is 0 so value is empty string: ''", bitstring.__class__.__name__)
        logger.info("%s: decoding finished.", bitstring.__class__.__name__)
        return ''
    if (bitstring.subtypeSpec.lowerEndpoint is not None) and (bitstring.subtypeSpec.lowerEndpoint == bitstring.subtypeSpec.upperEndpoint):
        logger.info("%s: lowerEndpoint == upperEndpoint == %s", bitstring.__class__.__name__, bitstring.subtypeSpec.lowerEndpoint)
        if bitstring.subtypeSpec.lowerEndpoint <= 16:    # 16.9
            bitstring_val = per_bytes.next(size=bit_field_size(octets=0, bits=bitstring.subtypeSpec.lowerEndpoint))
            bitstring_val = bin(bitstring_val)[2:].zfill(bitstring.subtypeSpec.lowerEndpoint)
            logger.info("%s: value: %s", bitstring.__class__.__name__, bitstring_val)
            logger.info("%s: decoding finished.", bitstring.__class__.__name__)
            return bitstring.__class__(bitstring_val)
        elif 16 < bitstring.subtypeSpec.lowerEndpoint < 65536:    # 16.10
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            bitstring_val = per_bytes.next(size=bit_field_size(octets=0, bits=bitstring.subtypeSpec.lowerEndpoint))
            bitstring_val = bin(bitstring_val)[2:].zfill(bitstring.subtypeSpec.lowerEndpoint)[:bitstring.subtypeSpec.lowerEndpoint]
            logger.info("%s: value: %s", bitstring.__class__.__name__, bitstring_val)
            logger.info("%s: decoding finished.", bitstring.__class__.__name__)
            return bitstring.__class__(bitstring_val)
        else:
            logger.error("%s: lowerEndpoint >= 65536 not supported", bitstring.__class__.__name__)
            raise NotImplemented
    else:    # 16.11
        if bitstring.subtypeSpec.upperEndpoint and bitstring.subtypeSpec.upperEndpoint < 65536:
            logger.info("%s: lowerEndpoint (%s) and upperEdnpoint (%s) < 65536", bitstring.__class__.__name__,
                        bitstring.subtypeSpec.lowerEndpoint, bitstring.subtypeSpec.upperEndpoint)
            bitstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                     constrained=True,
                                                                     n=per_bytes,
                                                                     lowerBound=bitstring.subtypeSpec.lowerEndpoint,
                                                                     upperBound=bitstring.subtypeSpec.upperEndpoint)
            bitstring_val = per_bytes.next(size=bit_field_size(octets=0, bits=bitstring_length_determinant), octet_align=True)
            bitstring_val = bin(bitstring_val)[2:].zfill(bitstring_length_determinant)
            logger.info("%s: value: %s", bitstring.__class__.__name__, bitstring_val)
            logger.info("%s: decoding finished.", bitstring.__class__.__name__)
            return bitstring.__class__(bitstring_val)
        else:
            logger.info("%s: lowerEndpoint = %s; upperEndpoint = %s", bitstring.__class__.__name__,
                        bitstring.subtypeSpec.lowerEndpoint, bitstring.subtypeSpec.upperEndpoint)
            bitstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                     constrained=False,
                                                                     n=per_bytes,
                                                                     lowerBound=0,
                                                                     upperBound=None)
            bitstring_val = per_bytes.next(size=bit_field_size(octets=0, bits=bitstring_length_determinant), octet_align=True)
            bitstring_val = bin(bitstring_val)[2:].zfill(bitstring_length_determinant)
            logger.info("%s: value: %s", bitstring.__class__.__name__, bitstring_val)
            logger.info("%s: decoding finished.", bitstring.__class__.__name__)
            return bitstring.__class__(bitstring_val)


def decode_octetstring(octetstring, per_bytes):
    logger.info("Decoding octetstring: %s", octetstring.__class__.__name__)
    if octetstring.subtypeSpec.extensionMarker:  # 17.3
        logger.info("%s: extension marker should be present", octetstring.__class__.__name__)
        extension_bit_field = per_bytes.next(size=bit_field_size(octets=0, bits=1))
        logger.info("%s: is octetstring value extened: %s", octetstring.__class__.__name__, extension_bit_field)
        if not extension_bit_field:
            pass
        else:
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            octetstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                       constrained=False,
                                                                       n=per_bytes,
                                                                       lowerBound=0,
                                                                       upperBound=None)
            if octetstring_length_determinant == 0:
                logger.info("%s: length determinant is 0 so value is empty string: ''", octetstring.__class__.__name__)
                logger.info("%s: decoding finished.", octetstring.__class__.__name__)
                return octetstring.__class__(octetstring.__class__.fromHexString(''))
            octetstring_val = per_bytes.next(size=bit_field_size(octets=octetstring_length_determinant, bits=0))
            octetstring_val = hex(octetstring_val)[2:].rstrip("L").zfill(octetstring_length_determinant * 2)
            logger.info("%s: value: %s", octetstring.__class__.__name__, octetstring_val)
            logger.info("%s: decoding finished.", octetstring.__class__.__name__)
            return octetstring.__class__(octetstring.__class__.fromHexString(octetstring_val))
    if octetstring.subtypeSpec.upperEndpoint == 0:  # 17.5
        logger.info("%s: upperEndpoint is 0 so value is empty string: ''", octetstring.__class__.__name__)
        logger.info("%s: decoding finished.", octetstring.__class__.__name__)
        return ''
    if (octetstring.subtypeSpec.lowerEndpoint is not None) and (octetstring.subtypeSpec.lowerEndpoint == octetstring.subtypeSpec.upperEndpoint):
        logger.info("%s: lowerEndpoint == upperEndpoint == %s", octetstring.__class__.__name__,
                    octetstring.subtypeSpec.lowerEndpoint)
        if octetstring.subtypeSpec.lowerEndpoint <= 2:    # 17.6
            octetstring_val = per_bytes.next(size=bit_field_size(octets=octetstring.subtypeSpec.lowerEndpoint, bits=0))
            octetstring_val = hex(octetstring_val)[2:].rstrip("L").zfill(octetstring.subtypeSpec.lowerEndpoint * 2)
            logger.info("%s: value: %s", octetstring.__class__.__name__, octetstring_val)
            logger.info("%s: decoding finished.", octetstring.__class__.__name__)
            return octetstring.__class__(octetstring.__class__.fromHexString(octetstring_val))
        elif 2 < octetstring.subtypeSpec.lowerEndpoint < 65536:  # 17.7
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            octetstring_val = per_bytes.next(size=bit_field_size(octets=octetstring.subtypeSpec.lowerEndpoint, bits=0))
            octetstring_val = hex(octetstring_val)[2:].rstrip("L").zfill(octetstring.subtypeSpec.lowerEndpoint * 2)
            logger.info("%s: value: %s", octetstring.__class__.__name__, octetstring_val)
            logger.info("%s: decoding finished.", octetstring.__class__.__name__)
            return octetstring.__class__(octetstring.__class__.fromHexString(octetstring_val))
        else:
            logger.error("%s: lowerEndpoint >= 65536 not supported", octetstring.__class__.__name__)
            raise NotImplemented
    else:  # 17.8
        logger.info("%s: lowerEndpoint = %s; upperEndpoint = %s", octetstring.__class__.__name__,
                    octetstring.subtypeSpec.lowerEndpoint, octetstring.subtypeSpec.upperEndpoint)
        if octetstring.subtypeSpec.upperEndpoint:
            octetstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                       constrained=True,
                                                                       n=per_bytes,
                                                                       lowerBound=octetstring.subtypeSpec.lowerEndpoint,
                                                                       upperBound=octetstring.subtypeSpec.upperEndpoint)
            if octetstring_length_determinant == 0:
                logger.info("%s: length determinant is 0 so value is empty string: ''", octetstring.__class__.__name__)
                logger.info("%s: decoding finished.", octetstring.__class__.__name__)
                return octetstring.__class__(octetstring.__class__.fromHexString(''))
            octetstring_val = per_bytes.next(size=bit_field_size(octets=octetstring_length_determinant, bits=0), octet_align=True)
            octetstring_val = hex(octetstring_val)[2:].rstrip("L")
            logger.info("%s: value: %s", octetstring.__class__.__name__, octetstring_val)
            logger.info("%s: decoding finished.", octetstring.__class__.__name__)
            return octetstring.__class__(octetstring.__class__.fromHexString(octetstring_val))
        else:
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            octetstring_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                       constrained=False,
                                                                       n=per_bytes,
                                                                       lowerBound=0,
                                                                       upperBound=None)
            octetstring_val = per_bytes.next(size=bit_field_size(octets=octetstring_length_determinant, bits=0))
            octetstring_val = hex(octetstring_val)[2:].rstrip("L").zfill(octetstring_length_determinant * 2)
            logger.info("%s: value: %s", octetstring.__class__.__name__, octetstring_val)
            logger.info("%s: decoding finished.", octetstring.__class__.__name__)
            return octetstring.__class__(octetstring.__class__.fromHexString(octetstring_val))


def decode_sequence(sequence, per_bytes):
    logger.info("Decoding sequence: %s", sequence.__class__.__name__)
    new_seq = sequence.__class__()
    extension_bit = 0
    optional_default_bit_field_len = 0
    optional_default_bit_field_value = 0
    if sequence.subtypeSpec.extensionMarker:    # 19.1
        logger.info("%s: extension marker should be present", sequence.__class__.__name__)
        extension_bit = per_bytes.next(size=bit_field_size(octets=0, bits=1))
    logger.info("%s: extension bit: %s", sequence.__class__.__name__, extension_bit)
    for namedType in sequence.rootComponent.namedTypes:    # 19.2
        if namedType.isOptional or namedType.isDefaulted:
            optional_default_bit_field_len += 1
    logger.info("%s: number of optional fields: %s", sequence.__class__.__name__, optional_default_bit_field_len)
    if optional_default_bit_field_len:
        if optional_default_bit_field_len < 65536:    # 19.3
            optional_default_bit_field_value = per_bytes.next(size=bit_field_size(octets=0, bits=optional_default_bit_field_len))
            optional_default_bit_field_value = bin(optional_default_bit_field_value)[2:].zfill(optional_default_bit_field_len)
            logger.info("%s: optional fields bits: %s", sequence.__class__.__name__, optional_default_bit_field_value)
            optional_default_bit_field_value = deque(optional_default_bit_field_value)
        else:
            logger.error("%s: optional_default_bit_field_len >= 65536 not supported.")
            raise NotImplemented
    for namedType in sequence.rootComponent.namedTypes:
        if namedType.isOptional:
            is_present = int(optional_default_bit_field_value.popleft())
            if is_present:
                logger.info("%s: decoding optional component: '%s' of type '%s' ...", sequence.__class__.__name__,
                            namedType.name, namedType.asn1Object.__class__.__name__)
                componentType = sequence.getComponentByName(namedType.name)
                decoded = componentType.create_field_list(per_bytes)
                logger.info("%s: decoded component: '%s'", sequence.__class__.__name__, namedType.name)
                new_seq[namedType.name] = decoded
            else:
                logger.info("%s: optional component '%s' not present", sequence.__class__.__name__, namedType.name)
                continue
        elif namedType.isDefaulted:    # CANONICAL-PER
            is_present = int(optional_default_bit_field_value.popleft())
            if is_present:
                logger.info("%s: decoding default component: '%s' of type '%s' ...", sequence.__class__.__name__,
                            namedType.name, namedType.asn1Object.__class__.__name__)
                componentType = sequence.getComponentByName(namedType.name)
                decoded = componentType.create_field_list(per_bytes)
                logger.info("%s: decoded component: '%s'", sequence.__class__.__name__, namedType.name)
                new_seq[namedType.name] = decoded
            else:
                logger.info("%s: default component '%s' not present", sequence.__class__.__name__, namedType.name)
                continue
        else:
            logger.info("%s: decoding component: '%s' of type '%s' ...", sequence.__class__.__name__,
                        namedType.name, namedType.asn1Object.__class__.__name__)
            componentType = sequence.getComponentByName(namedType.name)
            decoded = componentType.create_field_list(per_bytes)
            logger.info("%s: decoded component: '%s'", sequence.__class__.__name__, namedType.name)
            new_seq[namedType.name] = decoded
    if extension_bit:    # 19.7
        extension_addition_len = len(sequence.extensionAddition.namedTypes) if sequence.extensionAddition else 0
        extension_addition_group_len = len(sequence.extensionAdditionGroups)
        extension_addition_bit_field_len = extension_addition_len + extension_addition_group_len
        extension_addition_length_determinant = decode_length_determinant(normally_small_length=True,
                                                                          constrained=False,
                                                                          n=per_bytes,
                                                                          lowerBound=None,
                                                                          upperBound=None)
        extension_addition_bit_field_value = per_bytes.next(size=bit_field_size(octets=0, bits=extension_addition_length_determinant))
        extension_addition_bit_field_bits = bin(extension_addition_bit_field_value)[2:].zfill(extension_addition_bit_field_len)
        logger.info("%s: extension bits: '%s'; within that: extension addition bits: '%s'; extension addition group bits: '%s'",
                    sequence.__class__.__name__,
                    extension_addition_bit_field_bits,
                    extension_addition_bit_field_bits[:extension_addition_len],
                    extension_addition_bit_field_bits[extension_addition_len:])
        per_bytes.next(bit_field_size(octets=0, bits=0), octet_align=True)
        if sequence.extensionAddition:
            for extensionNamedType, is_present_bit in zip(sequence.extensionAddition.namedTypes,
                                                          extension_addition_bit_field_bits[:extension_addition_len]):
                if int(is_present_bit):
                    logger.info("%s: extension addition component '%s' present. Decoding...", sequence.__class__.__name__,
                                extensionNamedType.name)
                    componentType = sequence.getComponentByName(extensionNamedType.name)
                    decoded = decode_open_type_field(componentType, per_bytes)
                    logger.info("%s: decoded extension addition component: '%s'", sequence.__class__.__name__, extensionNamedType.name)
                    new_seq[extensionNamedType.name] = decoded
                else:
                    logger.info("%s: extension addition component '%s' not present.", sequence.__class__.__name__,
                                extensionNamedType.name)
        if sequence.extensionAdditionGroups:
            for group_index, (extensionAdditionGroup, is_present_bit) in enumerate(zip(sequence.extensionAdditionGroups,
                                                              extension_addition_bit_field_bits[extension_addition_len:]), 1):
                if int(is_present_bit):
                    logger.info("%s: extension addition group number %s present. Decoding as separate sequence with name 'INNER_SEQUENCE' ...",
                                sequence.__class__.__name__, group_index)
                    group_sequence = create_new_fake_empty_sequence(rootComponent=extensionAdditionGroup)
                    decoded = decode_open_type_field(group_sequence, per_bytes)
                    logger.info("%s: extension addition group decoded; Components: '%s'",
                                sequence.__class__.__name__,
                                ', '.join([type_name for type_name in decoded.named_types.keys()]))
                    for type_name, type_value in decoded.named_types.items():
                        new_seq[type_name] = type_value
                else:
                    logger.info("%s: extension addition group number %s not present",
                                sequence.__class__.__name__, group_index)
    logger.info("%s: decoding finished.", sequence.__class__.__name__)
    return new_seq


def decode_sequence_of(sequence_of, per_bytes):
    logger.info("Decoding sequence of: %s", sequence_of.__class__.__name__)
    new_seq_of = sequence_of.__class__()
    component_type = sequence_of.getComponentType()
    if sequence_of.subtypeSpec.extensionMarker:  # 20.4
        logger.info("%s: extension marker should be present", sequence_of.__class__.__name__)
        extension_bit_field = per_bytes.next(size=bit_field_size(octets=0, bits=1), octet_align=False)
        logger.info("%s: extension bit: %s", sequence_of.__class__.__name__, extension_bit_field)
        if not extension_bit_field:
            length_determinant = decode_length_determinant(normally_small_length=False,
                                                           constrained=True,
                                                           n=per_bytes,
                                                           lowerBound=sequence_of.subtypeSpec.lowerEndpoint,
                                                           upperBound=sequence_of.subtypeSpec.upperEndpoint)
            logger.info("%s: decoding %s component(s) of type: '%s'", sequence_of.__class__.__name__,
                        length_determinant, component_type.__class__.__name__)
            for index in range(length_determinant):
                val = component_type.create_field_list(per_bytes)
                logger.info("%s: decoding of component '%s' number %s finished", sequence_of.__class__.__name__,
                            component_type.__class__.__name__, index + 1)
                new_seq_of.extend([val])
            logger.info("%s: decoding finished.", sequence_of.__class__.__name__)
            return new_seq_of
        else:
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
            length_determinant = decode_length_determinant(normally_small_length=False,
                                                           constrained=False,
                                                           n=per_bytes,
                                                           lowerBound=0,
                                                           upperBound=None)
            logger.info("%s: decoding %s component(s) of type: '%s'", sequence_of.__class__.__name__,
                        length_determinant, component_type.__class__.__name__)
            for index in range(length_determinant):
                val = component_type.create_field_list(per_bytes)
                logger.info("%s: decoding of component '%s' number %s finished", sequence_of.__class__.__name__,
                            component_type.__class__.__name__, index + 1)
                new_seq_of.extend([val])
            logger.info("%s: decoding finished.", sequence_of.__class__.__name__)
            return new_seq_of
    elif (sequence_of.subtypeSpec.lowerEndpoint == sequence_of.subtypeSpec.upperEndpoint) and (sequence_of.subtypeSpec.upperEndpoint < 65536):    # 20.5
        logger.info("%s: lowerEndpoint == upperEndpoint and upperEndpoint < 65536")
        logger.info("%s: decoding %s component(s) of type: '%s'", sequence_of.__class__.__name__,
                    sequence_of.subtypeSpec.lowerEndpoint, component_type.__class__.__name__)
        for index in range(sequence_of.subtypeSpec.lowerEndpoint):
            val = component_type.create_field_list(per_bytes)
            logger.info("%s: decoding of component '%s' number %s finished", sequence_of.__class__.__name__,
                        component_type.__class__.__name__, index + 1)
            new_seq_of.extend([val])
        logger.info("%s: decoding finished.", sequence_of.__class__.__name__)
        return new_seq_of
    else:  # 20.6
        logger.info("%s: lowerEndpoint = %s and upperEndpoint = %s", sequence_of.__class__.__name__,
                    sequence_of.subtypeSpec.lowerEndpoint, sequence_of.subtypeSpec.upperEndpoint)
        if sequence_of.subtypeSpec.upperEndpoint:
            length_determinant = decode_length_determinant(normally_small_length=False,
                                                           constrained=True,
                                                           n=per_bytes,
                                                           lowerBound=sequence_of.subtypeSpec.lowerEndpoint,
                                                           upperBound=sequence_of.subtypeSpec.upperEndpoint)
            logger.info("%s: decoding %s component(s) of type: '%s'", sequence_of.__class__.__name__,
                        length_determinant, component_type.__class__.__name__)
            for index in range(length_determinant):
                val = component_type.create_field_list(per_bytes)
                logger.info("%s: decoding of component '%s' number %s finished", sequence_of.__class__.__name__,
                            component_type.__class__.__name__, index + 1)
                new_seq_of.extend([val])
            logger.info("%s: decoding finished.", sequence_of.__class__.__name__)
            return new_seq_of
        else:
            length_determinant = decode_length_determinant(normally_small_length=False,
                                                           constrained=False,
                                                           n=per_bytes,
                                                           lowerBound=sequence_of.subtypeSpec.lowerEndpoint,
                                                           upperBound=None)
            logger.info("%s: decoding %s component(s) of type: '%s'", sequence_of.__class__.__name__,
                        length_determinant, component_type.__class__.__name__)
            for index in range(length_determinant):
                val = component_type.create_field_list(per_bytes)
                logger.info("%s: decoding of component '%s' number %s finished", sequence_of.__class__.__name__,
                            component_type.__class__.__name__, index + 1)
                new_seq_of.extend([val])
            logger.info("%s: decoding finished.", sequence_of.__class__.__name__)
            return new_seq_of


def decode_choice(choice, per_bytes):    # 23
    logger.info("Decoding choice: %s", choice.__class__.__name__)
    new_choice = choice.__class__()
    n = len(choice.rootComponent.namedTypes) - 1
    if not choice.subtypeSpec.extensionMarker:  # 23.6
        if len(choice.rootComponent.namedTypes) == 1:  # 23.4
            logger.info("%s: length of choice list is 1", choice.__class__.__name__)
            selected_type = choice.rootComponent.namedTypes[0]
            selected_component = choice.getComponentByName(selected_type.name)
            logger.info("%s: chosen component: '%s' of type '%s' has index 0. Decoding...", choice.__class__.__name__,
                        selected_type.name, selected_component.__class__.__name__)
            val = selected_component.create_field_list(per_bytes)
            logger.info("%s: decoding of chosen component '%s' finished", choice.__class__.__name__, selected_type.name)
            new_choice[selected_type.name] = val
            logger.info("%s: decoding finished.", choice.__class__.__name__)
            return new_choice
        else:    # 23.6
            index = decode_constrained_whole_number_from_field_list(per_bytes, lowerEndpoint=0, upperEndpoint=n)
            namedType = choice.rootComponent.namedTypes[index]
            selected_component = choice.getComponentByPosition(index)
            logger.info("%s: chosen component: '%s' of type '%s' has index %s. Decoding...", choice.__class__.__name__,
                        namedType.name, selected_component.__class__.__name__, index)
            val = selected_component.create_field_list(per_bytes)
            logger.info("%s: decoding of chosen component '%s' finished", choice.__class__.__name__, namedType.name)
            new_choice[namedType.name] = val
            logger.info("%s: decoding finished.", choice.__class__.__name__)
            return new_choice
    else:    # 23.5
        logger.info("%s: extension marker should be present", choice.__class__.__name__)
        extension_bit = per_bytes.next(size=bit_field_size(octets=0, bits=1), octet_align=False)
        logger.info("%s: extension bit: %s", choice.__class__.__name__, extension_bit)
        if not extension_bit:    # 23.7
            if len(choice.rootComponent.namedTypes) == 1:  # 23.4
                logger.info("%s: length of choice list is 1", choice.__class__.__name__)
                selected_type = choice.rootComponent.namedTypes[0]
                selected_component = choice.getComponentByName(selected_type.name)
                logger.info("%s: chosen component: '%s' of type '%s' has index 0. Decoding...",
                            choice.__class__.__name__,
                            selected_type.name, selected_component.__class__.__name__)
                val = selected_component.create_field_list(per_bytes)
                logger.info("%s: decoding of chosen component '%s' finished", choice.__class__.__name__,
                            selected_type.name)
                new_choice[selected_type.name] = val
                logger.info("%s: decoding finished.", choice.__class__.__name__)
                return new_choice
            else:
                index = decode_constrained_whole_number_from_field_list(per_bytes, lowerEndpoint=0, upperEndpoint=n)
                namedType = choice.rootComponent.namedTypes[index]
                selected_component = choice.getComponentByPosition(index)
                logger.info("%s: chosen component: '%s' of type '%s' has index %s. Decoding...",
                            choice.__class__.__name__,
                            namedType.name, selected_component.__class__.__name__, index)
                val = selected_component.create_field_list(per_bytes)
                logger.info("%s: decoding of chosen component '%s' finished", choice.__class__.__name__, namedType.name)
                new_choice[namedType.name] = val
                logger.info("%s: decoding finished.", choice.__class__.__name__)
                return new_choice
        else:    # 23.8
            index = decode_normally_small_non_negative_whole_number(per_bytes, lower_bound=0)
            if choice.extensionAddition and not choice.extensionAdditionGroups:
                logger.info("%s: has only extension addition", choice.__class__.__name__)
                choice_extension = choice.extensionAddition
            elif not choice.extensionAddition and choice.extensionAdditionGroups:
                logger.info("%s: has only extension addition groups", choice.__class__.__name__)
                choice_extension = choice.extensionAdditionGroups[0].__class__(*[named_type for additive_named_types in choice.extensionAdditionGroups for
                                                                                 named_type in additive_named_types._NamedTypes__namedTypes])
            elif choice.extensionAddition and choice.extensionAdditionGroups:
                logger.info("%s: has extension addition and extension addition groups", choice.__class__.__name__)
                choice_extension = choice.extensionAddition + choice.extensionAdditionGroups
            else:  # is this a valid case?
                logger.error("%s: extension bit is present but neither extension addition nor extension addition groups are.",
                             choice.__class__.__name__)
                raise NotImplemented
            namedType = choice_extension.namedTypes[index]
            selected_component = choice.getComponentByName(namedType.name)
            logger.info("%s: chosen component: '%s' of type '%s' has index %s (indexed from extension choices). Decoding...",
                        choice.__class__.__name__,
                        namedType.name, selected_component.__class__.__name__, index)
            val = decode_open_type_field(selected_component, per_bytes)
            logger.info("%s: decoding of chosen component '%s' finished", choice.__class__.__name__, namedType.name)
            new_choice[namedType.name] = val
            logger.info("%s: decoding finished.", choice.__class__.__name__)
            return new_choice


def decode_constrained_whole_number_from_field_list(per_bytes, lowerEndpoint, upperEndpoint):
    logger.debug("function: decode_constrained_whole_number_from_field_list; params: per_bytes: (hidden), "
                 "lowerEndpoint: %s, upperEndpoint: %s", lowerEndpoint, upperEndpoint)
    if integer_range(lowerEndpoint, upperEndpoint) <= 65536:  # 13.2.5
        return decode_constrained_whole_number(per_bytes, lowerBound=lowerEndpoint, upperBound=upperEndpoint)
    else:    # 13.2.6
        upperBound = get_range_octet_len(upperEndpoint - lowerEndpoint)
        constrained_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                   constrained=True,
                                                                   n=per_bytes,
                                                                   lowerBound=1,
                                                                   upperBound=upperBound)

        val = per_bytes.next(bit_field_size(octets=constrained_length_determinant, bits=0), octet_align=True)
        return lowerEndpoint + val


def decode_semi_constrained_whole_number_from_field_list(per_bytes, lowerBound):
    logger.debug("function: decode_semi_constrained_whole_number_from_field_list; params: per_bytes: (hidden), lowerBound: %s",
                 lowerBound)
    unconstrained_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                 constrained=False,
                                                                 n=per_bytes,
                                                                 lowerBound=1,
                                                                 upperBound=None)
    logger.debug("semi constranied length determinant: %s", unconstrained_length_determinant)
    val = per_bytes.next(bit_field_size(octets=unconstrained_length_determinant, bits=0))
    logger.debug("semi constrained whole number value: %s", val)
    return lowerBound + val


def decode_open_type_field(type, per_bytes):
    unconstrained_length_determinant = decode_length_determinant(normally_small_length=False,
                                                                 constrained=False,
                                                                 n=per_bytes,
                                                                 lowerBound=None,
                                                                 upperBound=None)
    val = per_bytes.next(size=bit_field_size(octets=unconstrained_length_determinant, bits=0), octet_align=True)
    open_type_value = PerBytes(hex(val)[2:].rstrip("L").zfill(unconstrained_length_determinant * 2))
    decoded = type.create_field_list(open_type_value)
    return decoded


def decode_constrained_whole_number(per_bytes, lowerBound, upperBound):
    logger.debug("function: decode_constrained_whole_number; params: per_bytes: (hidden), lowerBound: %s, "
                 "upperBound: %s", lowerBound, upperBound)
    range = integer_range(lowerBound, upperBound)
    field_size = get_bit_field_size(range)
    logger.debug("constrained whole number: range: %s; size: %s", range, field_size)
    if range == 1:
        return 0
    elif range <= 255:
        non_negative_binary_integer = decode_non_negative_binary_integer(per_bytes, field_size, octet_align=False)
    elif range == 256:
        per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
        non_negative_binary_integer = decode_non_negative_binary_integer(per_bytes, field_size, octet_align=True)
    elif 257 <= range <= 65536:
        per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
        non_negative_binary_integer = decode_non_negative_binary_integer(per_bytes, field_size, octet_align=True)
    elif range > 65536:
        non_negative_binary_integer = decode_non_negative_binary_integer(per_bytes, field_size,
                                                                         octet_align=True,
                                                                         minimum_number_of_octets=True)
        return non_negative_binary_integer
    return lowerBound + non_negative_binary_integer


def decode_normally_small_non_negative_whole_number(per_bytes, lower_bound):    # 11.6
    logger.debug("function: decode_normally_small_non_negative_whole_number; params: per_bytes: (hidden), lowerBound: %s",
                 lower_bound)
    single_bit_field = per_bytes.next(size=bit_field_size(octets=0, bits=1))
    if not single_bit_field:    # 11.6.1    # n <= 63
        normally_small_non_negative_whole_number = decode_non_negative_binary_integer(per_bytes=per_bytes,
                                                                                      field_size=bit_field_size(octets=0, bits=6),
                                                                                      octet_align=False)
        return normally_small_non_negative_whole_number
    else:    # 11.6.2
        per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
        semi_constrained_whole_number = decode_semi_constrained_whole_number_from_field_list(per_bytes, lowerBound=lower_bound)
        return semi_constrained_whole_number


def decode_unconstrained_whole_number(per_bytes):
    logger.debug("function: decode_unconstrained_whole_number; params: per_bytes: (hidden)")
    unconstrained_length_determinant = per_bytes.next(octet_align=True)
    logger.debug("unconstarined length determinant: %s", unconstrained_length_determinant)
    unconstrained_whole_number = twos_comp(
        per_bytes=per_bytes.next(bit_field_size(octets=unconstrained_length_determinant, bits=0)),
        bits=8*unconstrained_length_determinant)
    logger.debug("unconstarined whole number: %s", unconstrained_whole_number)
    return unconstrained_whole_number


def decode_non_negative_binary_integer(per_bytes, field_size, octet_align, minimum_number_of_octets=False):
    logger.debug("function: decode_non_negative_binary_integer; params: per_bytes: (hidden), field_size: %s, octet_align: %s, "
                 "minimum_number_of_octets: %s", field_size, octet_align, minimum_number_of_octets)
    if minimum_number_of_octets:
        logger.error("Parameter 'minimum_number_of_octets' no supported.")
        raise NotImplemented
    else:
        decoded = per_bytes.next(field_size)
        logger.debug("Non-negative binary integer value: '%s'", decoded)
        if octet_align:
            per_bytes.next(size=bit_field_size(octets=0, bits=0), octet_align=octet_align)
        return decoded


def twos_comp(per_bytes, bits):
    """compute the 2's complement of int value val"""
    if (per_bytes & (1 << (bits - 1))) != 0: # if sign bit is set e.g., 8bit: 128-255
        per_bytes = per_bytes - (1 << bits)        # compute negative value
    return per_bytes


def decode_length_determinant(normally_small_length, constrained, n, lowerBound, upperBound):
    logger.debug("function: decode_length_determinant; params: normally_small_length: %s, constrained: %s, "
                 "n (per_bytes): (hidden), lowerBound: %s, upperBound: %s", normally_small_length, constrained, lowerBound,
                 upperBound)
    if normally_small_length:    # 11.9.3.4
        single_bit_field = n.next(size=bit_field_size(octets=0, bits=1))
        if not single_bit_field:    # n <= 64
            normally_small_length_determinant_size = decode_non_negative_binary_integer(per_bytes=n,
                                                                                        field_size=bit_field_size(octets=0, bits=6),
                                                                                        octet_align=False)
            return normally_small_length_determinant_size + 1
        else:    # n > 64
            raise NotImplemented
    elif constrained:
        constrained_length_determinant = decode_constrained_whole_number(per_bytes=n, lowerBound=lowerBound, upperBound=upperBound)
        return constrained_length_determinant
    else:
        n.next(size=bit_field_size(octets=0, bits=0), octet_align=True)
        first_bit = n.next(bit_field_size(octets=0, bits=1))
        if not first_bit:    # 11.9.3.6
            unconstrained_length_determinant = decode_non_negative_binary_integer(per_bytes=n,
                                                                                  field_size=bit_field_size(octets=0, bits=7),
                                                                                  octet_align=True)
            return unconstrained_length_determinant
        else:
            second_bit = n.next(bit_field_size(octets=0, bits=1))
            if not second_bit:    # 11.9.3.7
                unconstrained_length_determinant = decode_non_negative_binary_integer(per_bytes=n,
                                                                                      field_size=bit_field_size(
                                                                                      octets=1, bits=6),
                                                                                      octet_align=True)
                return unconstrained_length_determinant
            else:
                raise NotImplemented


def create_new_fake_empty_sequence(rootComponent):
    class FooConstraint(object):
        pass

    class INNER_SEQUENCE(object):
        subtypeSpec = FooConstraint()
        def __init__(self):
            self.named_types = {}

        def create_field_list(self, per_bytes):
            return decode_sequence(self, per_bytes)

        def getComponentByName(self, name):
            idx = self.rootComponent.getPositionByName(name)
            return self.rootComponent.getTypeByPosition(idx)

        def __setitem__(self, key, value):
            self.named_types[key] = value

    INNER_SEQUENCE.subtypeSpec.extensionMarker = False
    INNER_SEQUENCE.subtypeSpec.lowerEndpoint= None
    INNER_SEQUENCE.subtypeSpec.upperEndpoint= None
    INNER_SEQUENCE.rootComponent = rootComponent
    INNER_SEQUENCE.componentType = rootComponent
    INNER_SEQUENCE.extensionAddition = None
    INNER_SEQUENCE.extensionAdditionGroups = []
    return INNER_SEQUENCE()



from pyparsing import oneOf, Word, ZeroOrMore, Regex, Literal, Empty
from .character_and_word_set import A_to_Z_latin, a_to_z_latin, \
    digits, hyphen_minus, non_breaking_hyphen, allowed_character_set, newline


'''
Rec. ITU-T X.680 (08/2015)
12.2 Type references
Page 17
'''
typereference = Word(A_to_Z_latin,
                     A_to_Z_latin + a_to_z_latin + digits +
                     hyphen_minus + non_breaking_hyphen)


'''
Rec. ITU-T X.680 (08/2015)
12.3 Identifiers
Page 17
'''
identifier = Word(a_to_z_latin,
                  a_to_z_latin + A_to_Z_latin + digits + hyphen_minus+ non_breaking_hyphen)


'''
Rec. ITU-T X.680 (08/2015)
12.4 Value references
Page 17
'''
valuereference = identifier

'''
Rec. ITU-T X.680 (08/2015)
12.5 Module references
Page 17
'''
modulereference = typereference


'''
Rec. ITU-T X.680 (08/2015)
12.6 Comments
Page 17
'''
# Missing comment for '/*' and '*/'
comment = hyphen_minus + \
          hyphen_minus + \
          ZeroOrMore(Word(allowed_character_set, allowed_character_set)) + \
          oneOf(hyphen_minus+ \
                hyphen_minus,
                newline)

'''
Rec. ITU-T X.680 (08/2015)
12.7 Empty lexical item
Page 18
'''
empty = Empty().suppress()

'''
Rec. ITU-T X.680 (08/2015)
12.8 Numbers
Page 18
'''
# number = Word(digits.replace('0', ''), digits, min=2) | \
#          Word(digits, max=1)
number = Regex(r'\d+')

'''
Rec. ITU-T X.680 (08/2015)
12.9 Real numbers
Page 18
'''
# This is a poor reflection of its definition but for time being will be enough
# realnumber = Word(digits, '.' + digits+ 'eE-')
realnumber = Regex(r'\d+\.\d+')

'''
Rec. ITU-T X.680 (08/2015)
12.10 Binary strings
Page 18
'''
# This is a poor reflection of its definition but for time being will be enough
# bstring = Word("'", "01B'")
bstring = Regex(r"\'[01\s]*\'B")


'''
Rec. ITU-T X.680 (08/2015)
12.12 Hexadecimal strings
Page 18/19
'''
# This is a poor reflection of its definition but for time being will be enough
# hstring = Word("'", digits + "ABCDEF'H")
hstring = Regex(r"\'[A-F0-9\s]*\'H")


'''
Rec. ITU-T X.680 (08/2015)
12.14 Character strings
Page 19
'''
# This is a poor reflection of its definition but for time being will be enough
# cstring = Word('"', digits + A_to_Z_latin + a_to_z_latin + '"')
cstring = Regex(r'\"[0-9A-Za-z\s]*\"')


'''
Rec. ITU-T X.680 (08/2015)
12.16 The simple character string lexical item
Page 22
'''
# This is a poor reflection of its definition but for time being will be enough
simplestring = cstring


'''
Rec. ITU-T X.680 (08/2015)
12.17 Time value character strings
Page 22
'''
tstring = Word('"', '0123456789+-:.,/CDHMRPSTWYZ"')


'''
Rec. ITU-T X.680 (08/2015)
12.19 The property and setting names lexical item
Page 22
'''
psname = simplestring


'''
Rec. ITU-T X.680 (08/2015)
12.20 Assignment lexical item
Page 22
'''
assignment = Literal('::=')


'''
Rec. ITU-T X.680 (08/2015)
12.21 Range separator
Page 22
'''
rangeseparator = Literal('..')


'''
Rec. ITU-T X.680 (08/2015)
12.22 Ellipsis
Page 22
'''
ellipsis = Literal('...')


'''
Rec. ITU-T X.680 (08/2015)
12.23 Left version brackets
Page 23
'''
leftversionbrackets = Literal('[[')


'''
Rec. ITU-T X.680 (08/2015)
12.24 Right version brackets
Page 23
'''
rightversionbrackets = Literal(']]')


'''
Rec. ITU-T X.680 (08/2015)
12.25 Encoding references
Page 23
'''
encodingreference = Word(A_to_Z_latin,
                    A_to_Z_latin + hyphen_minus+ non_breaking_hyphen)


# 12.37 Single character lexical items
LEFT_CURLY_BRACKET = Literal('{').suppress()
RIGHT_CURLY_BRACKET = Literal('}').suppress()
LESS_THAN_SIGN = Literal('<')
GREATER_THAN_SIGN = Literal('>')
COMMA = Literal(',').suppress()
FULL_STOP = Literal('.')
SOLIDUS = Literal('/')
LEFT_PARENTHESIS = Literal('(').suppress()
RIGHT_PARENTHESIS = Literal(')').suppress()
LEFT_SQUARE_BRACKET = Literal('[')
RIGHT_SQUARE_BRACKET = Literal(']')
HYPHEN_MINUS = Literal('-')
COLON = Literal(':')
EQUALS_SIGN = Literal('=')
QUOTATION_MARK = Literal('"')
APOSTROPHE = Literal("'")
SPACE = Literal(' ')
SEMICOLON = Literal(';')
COMMERCIAL_AT = Literal('@')
VERTICAL_LINE = Literal('|')
EXCLAMATION_MARK = Literal('!')
CIRCUMFLEX_ACCENT = Literal('^')


if __name__ == '__main__':
    s = "000'B"
    print(bstring.parseString(s, parseAll=True))


from pyparsing import Or, Forward, Literal, ZeroOrMore
from .common_def import SignedNumber, DefinedValue, NULL, CharacterStringList, Quadruple, Tuple
from .lexical_items import identifier, realnumber, HYPHEN_MINUS, COMMA, \
    LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, bstring, hstring, cstring


TRUE = Literal("TRUE")
FALSE = Literal("FALSE")
PLUS_INFINITY = Literal("PLUS-INFINITY")
MINUS_INFINITY = Literal("MINUS-INFINITY")
NOT_A_NUMBER = Literal("NOT-A-NUMBER")
CONTAINING = Literal("CONTAINING")


Value = Forward()
# ComponentValueList = Forward()


# 17.13
NamedValue = identifier + Value


# 25.18
ComponentValueList = Or([
    NamedValue + ZeroOrMore(COMMA + NamedValue),
    # NamedValue,
    # ComponentValueList + COMMA + NamedValue    # ORIGINAL
])


# 25.18
SequenceValue = Or([
    LEFT_CURLY_BRACKET + ComponentValueList + RIGHT_CURLY_BRACKET,
    LEFT_CURLY_BRACKET + RIGHT_CURLY_BRACKET
])


# 21.6
SpecialRealValue = Or([
    PLUS_INFINITY,
    MINUS_INFINITY,
    NOT_A_NUMBER
])


# 21.6
NumericRealValue = Or([
    realnumber,
    HYPHEN_MINUS + realnumber,
    SequenceValue
])


# 21.6
RealValue = Or([
    NumericRealValue,
    SpecialRealValue
])


# 17.11
ReferencedValue = Or([
    DefinedValue,
    # ValueFromObject    # Rec. ITU-T X.681 | ISO/IEC 8824-2, clause 15
])


# 19.9
IntegerValue = Or([
    SignedNumber,
    identifier
])


# 23.3
OctetStringValue = Or([
    bstring,
    hstring,
    CONTAINING + Value
])


# 18.3
BooleanValue = Or([
    TRUE,
    FALSE
])


# 24.3
NullValue = NULL


# 41.8
RestrictedCharacterStringValue = Or([
    cstring,
    CharacterStringList,
    Quadruple,
    Tuple
])


# 40.3
CharacterStringValue = Or([
    RestrictedCharacterStringValue,
    # UnrestrictedCharacterStringValue,
])


# 17.9
BuiltinValue = Or([
    # BitStringValue,
    BooleanValue,
    CharacterStringValue,
    # ChoiceValue,
    # EmbeddedPDVValue,
    # EnumeratedValue,
    # ExternalValue,
    # InstanceOfValue,
    IntegerValue,
    # IRIValue,
    NullValue,
    # ObjectIdentifierValue,
    OctetStringValue,
    RealValue,
    # RelativeIRIValue,
    # RelativeOIDValue,
    SequenceValue,
    # SequenceOfValue,
    # SetValue,
    # SetOfValue,
    # PrefixedValue,
    # TimeValue
])


# 17.7
Value << Or([
    BuiltinValue,
    ReferencedValue,
    # ObjectClassFieldValue    # Rec. ITU-T X.681 | ISO/IEC 8824-2, 14.6.
])



from asn1PERser.test.parsing.asn1_python_code.SimpleProtocol import *
from asn1PERser.codec.per.encoder import encode


def DATA_alive_message():
    '''
    simple_message SimpleMessage ::= alive : {
        timestamp {
            seconds 1557528149,
            useconds 12345
        }
    }
    '''
    utc_timestamp = UTC_Timestamp()
    utc_timestamp['seconds'] = UTC_Timestamp.seconds(1557528149)
    utc_timestamp['useconds'] = UTC_Timestamp.useconds(12345)

    msg_alive = Alive()
    msg_alive['timestamp'] = utc_timestamp

    simple_message = SimpleMessage()
    simple_message['alive'] = msg_alive

    return simple_message


def DATA_start_message():
    '''
    simple_message SimpleMessage ::= start : {
        sequenceNumber    10,
        timestamp {
            seconds 1557528149,
            useconds 12345
        },
        srcPort    65533,
        dstPort    10000
    }
    '''
    utc_timestamp = UTC_Timestamp()
    utc_timestamp['seconds'] = UTC_Timestamp.seconds(1557528149)
    utc_timestamp['useconds'] = UTC_Timestamp.useconds(12345)

    msg_start = Start()
    msg_start['sequenceNumber'] = SequenceNumber(10)
    msg_start['timestamp'] = utc_timestamp
    msg_start['srcPort'] = Port(65533)
    msg_start['dstPort'] = Port(10000)

    simple_message = SimpleMessage()
    simple_message['start'] = msg_start

    return simple_message


def DATA_data_message_1():
    '''
    simple_message SimpleMessage ::= data : {
        sequenceNumber    55555,
        swRelease  rel2,
        macroId    '11110000111100001111'B,
        payload {
            'DEAD'H,
            'BEEF'H,
            'FEED'H,
            'AA'H,
            'BBBBBBBB'H
        }
    }
    '''
    data_payload = Payload()
    data_payload.extend([Message(hexValue='DEAD')])
    data_payload.extend([Message(hexValue='BEEF')])
    data_payload.extend([Message(hexValue='FEED')])
    data_payload.extend([Message(hexValue='AA')])
    data_payload.extend([Message(hexValue='BBBBBBBB')])

    msg_data = Data()
    msg_data['sequenceNumber'] = SequenceNumber(55555)
    msg_data['swRelease'] = Data.swRelease('rel2')
    msg_data['macroId'] = Data.macroId(binValue='11110000111100001111')
    msg_data['payload'] = data_payload

    simple_message = SimpleMessage()
    simple_message['data'] = msg_data

    return simple_message


def DATA_data_message_2():
    '''
    simple_message SimpleMessage ::= data : {
        sequenceNumber    256,
        swRelease  rel3,
        payload {
            'DEADBEEF'H
        }
    }
    '''
    data_payload = Payload()
    data_payload.extend([Message(hexValue='DEADBEEF')])

    msg_data = Data()
    msg_data['sequenceNumber'] = SequenceNumber(256)
    msg_data['swRelease'] = Data.swRelease('rel3')
    msg_data['payload'] = data_payload

    simple_message = SimpleMessage()
    simple_message['data'] = msg_data

    return simple_message


def test_alive_message_can_be_encoded():
    per_encoded = encode(asn1Spec=DATA_alive_message())
    # assert per_encoded == bytes.fromhex('4C5CD5FE55403039')
    assert per_encoded == bytearray.fromhex('4C5CD5FE55403039')


def test_start_message_can_be_encoded():
    per_encoded = encode(asn1Spec=DATA_start_message())
    assert per_encoded == bytearray.fromhex('00000AC05CD5FE55403039D8ED0000')


def test_data_message_1_can_be_encoded():
    per_encoded = encode(asn1Spec=DATA_data_message_1())
    assert per_encoded == bytearray.fromhex('70D90320F0F0F880DEAD40BEEF40FEED00AAC0BBBBBBBB')


def test_data_message_2_can_be_encoded():
    per_encoded = encode(asn1Spec=DATA_data_message_2())
    assert per_encoded == bytearray.fromhex('60010043DEADBEEF')



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


minNumber = IntegerType(-10)


maxNumber = IntegerType(255)


class MyEnum(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('one', 0),
        ('two', 1),
        ('three', 2),
    )
    namedValues = enumerationRoot


class MySeq1(SequenceType):
    class my_Int1(IntegerType):
        subtypeSpec = ValueRange(minNumber, 10, extensionMarker=True)

    class myBitString1(BitStringType):
        subtypeSpec = ValueSize(0, maxNumber)

    class myOctetString1(OctetStringType):
        subtypeSpec = ValueSize(1, 4)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('my-Int1', my_Int1()),
        NamedType('my-Int2', IntegerType()),
        NamedType('myBitString1', myBitString1()),
        NamedType('myOctetString1', myOctetString1()),
    )
    componentType = rootComponent


class MySeq2(SequenceType):
    class myOctetString2(OctetStringType):
        subtypeSpec = ValueSize(1, 4, extensionMarker=True)

    class myInt4(IntegerType):
        subtypeSpec = ValueRange(MIN, 1)

    rootComponent = AdditiveNamedTypes(
        NamedType('myInt3', IntegerType()),
        NamedType('myOctetString2', myOctetString2()),
        NamedType('myInt4', myInt4()),
    )
    componentType = rootComponent


class Data(SequenceType):
    class i0(IntegerType):
        subtypeSpec = ValueRange(0, MAX)

    class i2(OctetStringType):
        subtypeSpec = ValueSize(1, 10)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('i0', i0()),
        NamedType('i1', OctetStringType()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('i2', i2()),
    )
    componentType = rootComponent + extensionAddition


class DataList(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, 10)
    componentType = Data()


class MyChoice(ChoiceType):
    class c0(OctetStringType):
        subtypeSpec = ValueSize(0, maxNumber, extensionMarker=True)

    class c_two(IntegerType):
        subtypeSpec = ValueRange(MIN, 0)

    class c4(BitStringType):
        subtypeSpec = ValueSize(0, 20)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('c0', c0()),
        NamedType('c1', IntegerType()),
        NamedType('c-two', c_two()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('c3', MyEnum()),
        NamedType('c4', c4()),
    )
    componentType = rootComponent + extensionAddition





from string import ascii_uppercase, ascii_lowercase, digits
from pyparsing import oneOf
from builtins import chr


RESERVED_WORDS = \
['ABSENT',
 'ENCODED',
 'INTERSECTION',
 'SEQUENCE',
 'ABSTRACT-SYNTAX',
 'ENCODING-CONTROL',
 'ISO646String',
 'SET',
 'ALL',
 'END',
 'MAX',
 'SETTINGS',
 'APPLICATION',
 'ENUMERATED',
 'MIN',
 'SIZE',
 'AUTOMATIC',
 'EXCEPT',
 'MINUS-INFINITY',
 'STRING',
 'BEGIN',
 'EXPLICIT',
 'NOT-A-NUMBER',
 'SYNTAX',
 'BIT',
 'EXPORTS',
 'NULL',
 'T61String',
 'BMPString',
 'EXTENSIBILITY',
 'NumericString',
 'TAGS',
 'BOOLEAN',
 'EXTERNAL',
 'OBJECT',
 'TeletexString',
 'BY',
 'FALSE',
 'ObjectDescriptor',
 'TIME',
 'CHARACTER',
 'FROM',
 'OCTET',
 'TIME-OF-DAY',
 'CHOICE',
 'GeneralizedTime',
 'OF',
 'TRUE',
 'CLASS',
 'GeneralString',
 'OID-IRI',
 'TYPE-IDENTIFIER',
 'COMPONENT',
 'GraphicString',
 'OPTIONAL',
 'UNION',
 'COMPONENTS',
 'IA5String',
 'PATTERN',
 'UNIQUE',
 'CONSTRAINED',
 'IDENTIFIER',
 'PDV',
 'UNIVERSAL',
 'CONTAINING',
 'IMPLICIT',
 'PLUS-INFINITY',
 'UniversalString',
 'DATE',
 'IMPLIED',
 'PRESENT',
 'UTCTime',
 'DATE-TIME',
 'IMPORTS',
 'PrintableString',
 'UTF8String',
 'DEFAULT',
 'INCLUDES',
 'PRIVATE',
 'VideotexString',
 'DEFINITIONS',
 'INSTANCE',
 'REAL',
 'VisibleString',
 'DURATION',
 'INSTRUCTIONS',
 'RELATIVE-OID',
 'WITH',
 'EMBEDDED',
 'INTEGER',
 'RELATIVE-OID-IRI']


'''
Rec. ITU-T X.680 (08/2015)
Table 2 - ASN.1 characters
Page 15
'''
A_to_Z_latin = ascii_uppercase
a_to_z_latin = ascii_lowercase
digits = digits
exclamation_mark = '!'
quotation_mark = '"'
ampersand = '&'
apostrophe = "'"
left_parenthesis = '('
right_parenthesis = ')'
asterisk = '*'
comma = ','
hyphen_minus = '-'
full_stop = '.'
solidus = '/'
colon = ':'
semicolon = ';'
less_than_sign = '<'
equals_sign = '='
greater_than_sign = '>'
commercial_at = '@'
left_square_bracket = '['
right_square_bracket = ']'
circumflex_accent = '^'
low_line = '_'
left_curly_bracket = '{'
vertical_line = '|'
right_curly_bracket = '}'
non_breaking_hyphen = chr(8211)

allowed_character_set = A_to_Z_latin + \
                        a_to_z_latin + \
                        digits + \
                        exclamation_mark + \
                        quotation_mark + \
                        ampersand + \
                        apostrophe + \
                        left_parenthesis + \
                        right_parenthesis + \
                        asterisk + \
                        comma + \
                        hyphen_minus + \
                        full_stop + \
                        solidus + \
                        colon + \
                        semicolon + \
                        less_than_sign + \
                        equals_sign + \
                        greater_than_sign + \
                        commercial_at + \
                        left_square_bracket + \
                        right_square_bracket + \
                        circumflex_accent + \
                        low_line + \
                        left_curly_bracket + \
                        vertical_line + \
                        right_curly_bracket + \
                        non_breaking_hyphen

'''
Rec. ITU-T X.680 (08/2015)
12.1.6
Page 16/17
'''
horizontal_tabulation = '\t'
line_feed = '\n'
vertical_tabulation = chr(11)
form_feed = chr(12)
carriage_return = '\r'
space = ' '
no_break_space = chr(255)

white_space = oneOf(horizontal_tabulation + \
                    line_feed + \
                    vertical_tabulation + \
                    form_feed + \
                    carriage_return + \
                    space + \
                    no_break_space
                    )

newline = oneOf(line_feed + \
                vertical_tabulation + \
                form_feed + \
                carriage_return
                )



from asn1PERser.codec.per.decoder import decode
from asn1PERser.codec.per.encoder import encode
from asn1PERser.classes.module import remove_comments
from asn1PERser.asn_definitions.module_def import ModuleDefinition
from asn1PERser.logging_setup import logging_setup


__all__ = ['decode', 'encode', 'parse_asn1_schema']

logging_setup()

def parse_asn1_schema(asn1_schema, output_folder=None):
    no_comment_schema = remove_comments(asn1_schema)
    parsed_asn = ModuleDefinition.parseString(no_comment_schema)[0]
    return parsed_asn.create_python_template(path=output_folder)



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class MyInteger(IntegerType):
    pass


class MyOctetString(OctetStringType):
    pass


class MyBitString(BitStringType):
    pass


class MyBoolean(BooleanType):
    pass


class MyEnumerated(EnumeratedType):
    enumerationRoot = NamedValues(
        ('four', 0),
        ('five', 1),
        ('six', 2),
    )
    namedValues = enumerationRoot


class MySeq(SequenceType):
    class d24(EnumeratedType):
        enumerationRoot = NamedValues(
            ('one', 0),
            ('two', 1),
            ('three', 2),
        )
        namedValues = enumerationRoot

    rootComponent = AdditiveNamedTypes(
        DefaultedNamedType('d00', IntegerType(10)),
        DefaultedNamedType('d01', IntegerType(0)),
        DefaultedNamedType('d02', MyInteger(11)),
        DefaultedNamedType('d03', MyInteger(0)),
        DefaultedNamedType('d04', OctetStringType(binValue='11111111')),
        DefaultedNamedType('d05', OctetStringType(binValue='')),
        DefaultedNamedType('d06', OctetStringType(hexValue='AAAA')),
        DefaultedNamedType('d07', OctetStringType(hexValue='')),
        DefaultedNamedType('d08', MyOctetString(binValue='01010101')),
        DefaultedNamedType('d09', MyOctetString(binValue='')),
        DefaultedNamedType('d10', MyOctetString(hexValue='BBBB')),
        DefaultedNamedType('d11', MyOctetString(hexValue='')),
        DefaultedNamedType('d12', BitStringType(binValue='0000')),
        DefaultedNamedType('d13', BitStringType(binValue='')),
        DefaultedNamedType('d14', BitStringType(hexValue='DDDD')),
        DefaultedNamedType('d15', BitStringType(hexValue='')),
        DefaultedNamedType('d16', MyBitString(binValue='10101')),
        DefaultedNamedType('d17', MyBitString(binValue='')),
        DefaultedNamedType('d18', MyBitString(hexValue='EE')),
        DefaultedNamedType('d19', MyBitString(hexValue='')),
        DefaultedNamedType('d20', BooleanType(False)),
        DefaultedNamedType('d21', BooleanType(True)),
        DefaultedNamedType('d22', MyBoolean(False)),
        DefaultedNamedType('d23', MyBoolean(True)),
        DefaultedNamedType('d24', d24('two')),
        DefaultedNamedType('d25', MyEnumerated('six')),
    )
    componentType = rootComponent





import os
import pytest
from asn1PERser.asn_definitions.module_def import ModuleDefinition


@pytest.fixture(scope='function')
def asn1_schema(request):
    schema_dir = os.path.split(__file__)[0]
    with open(os.path.join(schema_dir, 'asn1_schemas/{name}'.format(name=request.param)), 'r') as schema:
        return schema.read()


@pytest.fixture(scope='function')
def python_code(request):
    parsed_dir = os.path.split(__file__)[0]
    with open(os.path.join(parsed_dir, 'asn1_python_code/{name}'.format(name=request.param)), 'r') as parsed_schema:
        return parsed_schema.read()


@pytest.fixture()
def parse_schema():
    def _parse_schema(schema):
        parsed = ModuleDefinition.parseString(schema)
        parsed_to_python = parsed[0].create_python_template(path=None)
        return parsed_to_python
    return _parse_schema



from .BooleanType import BooleanType
from .SequenceType import SequenceType
from .UTF8String import UTF8String
from .IntegerType import IntegerType
from .EnumeratedType import EnumeratedType
from .ChoiceType import ChoiceType
from .SequenceOfType import SequenceOfType
from .BitStringType import BitStringType
from .OctetStringType import OctetStringType



from pyasn1.type.univ import Enumerated
from asn1PERser.codec.per.encoder import encode_enumerated
from asn1PERser.codec.per.decoder import decode_enumerated
from asn1PERser.classes.types.constraint import NoConstraint


class EnumeratedType(Enumerated):
    subtypeSpec = NoConstraint()
    enumerationRoot = None
    extensionAddition = None

    def fill_field_list(self, field_list):
        type_field_list = encode_enumerated(self)
        if type_field_list:
            field_list.extend(type_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_enumerated(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        return self.namedValues[int(self)]
